#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test :any:`LightObject`."""

import re

from hypothesis import assume, given
from hypothesis import strategies as st
from pydantic import ValidationError
from pytest import raises

import ucdp as u


class MyLightObject(u.LightObject):
    """Example LightObject."""

    arg1: int


class MyMyLightObject(MyLightObject):
    """Example Sub-LightObject."""

    arg2: int


class MyTuplingLightObject(u.LightObject):
    """Example LightObject with Tuple."""

    arg1: tuple[int, ...] = u.Field(default_factory=tuple)


class MyListingLightObject(u.LightObject):
    """Example LightObject with List."""

    arg1: list[int] = u.Field(default_factory=list)


@given(int1=st.integers(max_value=100), int2=st.integers(max_value=100), str1=st.text(max_size=10))
def test_light_object_basics(int1, int2, str1):
    """:any:`LightObject` Basic Testing."""
    inst = MyLightObject(arg1=int1)
    assert inst.arg1 == int1

    # compare
    assert MyLightObject(arg1=int1) == MyLightObject(arg1=int1)

    # caching
    assert MyLightObject(arg1=int1) is MyLightObject(arg1=int1)

    assert str(MyLightObject(arg1=int1)) == f"MyLightObject(arg1={int1!r})"
    assert repr(MyLightObject(arg1=int1)) == f"MyLightObject(arg1={int1!r})"

    # immutable
    with raises(ValidationError):
        inst.arg1 = int2

    # no extra
    with raises(ValidationError):
        MyLightObject(arg1=int1, arg2=int2)

    # validate
    with raises(ValidationError):
        MyLightObject(arg1=str1)

    assert hash(inst) is not None


@given(
    int1=st.integers(max_value=100),
    int2=st.integers(max_value=100),
    int3=st.integers(max_value=100),
    str1=st.text(max_size=10),
)
def test_sub_light_object(int1, int2, int3, str1):
    """Sub :any:`LightObject` Testing."""
    inst = MyMyLightObject(arg1=int1, arg2=int2)
    assert inst.arg1 == int1
    assert inst.arg2 == int2

    with raises(ValidationError):
        inst.arg1 = int3

    with raises(ValidationError):
        MyMyLightObject(arg1=int1, arg2=str1)

    assume(int2 != int3)

    # caching
    assert MyMyLightObject(arg1=int1, arg2=int2) is MyMyLightObject(arg1=int1, arg2=int2)
    assert MyMyLightObject(arg1=int1, arg2=int2) is MyMyLightObject(arg2=int2, arg1=int1)
    assert MyMyLightObject(arg1=int1, arg2=int2) is not MyMyLightObject(arg1=int1, arg2=int3)
    assert MyMyLightObject(arg1=int1, arg2=int3) is MyMyLightObject(arg1=int1, arg2=int3)


@given(int1=st.integers(max_value=100), int2=st.integers(max_value=100), tuple1=st.tuples(st.integers(max_value=100)))
def test_tupling_light_object(int1, int2, tuple1):
    """:any:`LightObject` With Tuple Testing."""
    assert MyTuplingLightObject().arg1 == ()

    with raises(ValidationError):
        MyTuplingLightObject(arg1=int1)

    inst = MyTuplingLightObject(arg1=tuple1)
    assert inst.arg1 == tuple1

    assert hash(inst) is not None


@given(int1=st.integers(max_value=100), list1=st.lists(st.integers(max_value=100)))
def test_listing_light_object(int1, list1):
    """:any:`LightObject` With List Testing."""
    with raises(TypeError):
        hash(MyListingLightObject())

    with raises(TypeError):
        MyListingLightObject(arg1=list1)

    with raises(ValidationError):
        MyListingLightObject(arg1=int1)


class NotHashableArg(u.LightObject):
    """Example."""

    arg: list[str]

    _posargs: u.PosArgs = ("arg",)

    def __init__(self, arg: list[str]):
        super().__init__(arg=arg)  # type: ignore[call-arg]


def test_not_hashable_arg():
    """Test Error Message on Non-Hashable Arg."""
    # TESTME

    #    NotHashableArg(["a", "b"])


class NotHashableKwarg(u.LightObject):
    """Example."""

    arg: list[str]


def test_not_hashable_kwarg():
    """Test Error Message on Non-Hashable Kwarg."""
    # TESTME

    #    NotHashableKwarg(["a", "b"])


class NoHashLightObject(u.LightObject):
    """Example LightObject."""

    arg1: list[int]

    def __init__(self, arg1):
        super().__init__(arg1=arg1)


def test_no_hash():
    """Not Hashable."""
    msg = "<class 'tests.test_light_object.NoHashLightObject'>: 0 argument [1, 2, 3] is not constant."
    with raises(TypeError, match=re.escape(msg)):
        NoHashLightObject([1, 2, 3])

    msg = "<class 'tests.test_light_object.NoHashLightObject'>: 'arg1' argument [1, 2, 3] is not constant."
    with raises(TypeError, match=re.escape(msg)):
        NoHashLightObject(arg1=[1, 2, 3])



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Tailored Module.

"""

from abc import abstractmethod
from typing import Any, ClassVar

from ._modbuilder import build
from .modbase import BaseMod
from .modfilelist import ModFileLists
from .modutil import get_libpath, get_topmodname, is_tb_from_modname
from .nameutil import join_names
from .object import PrivateField


class ATailoredMod(BaseMod):
    """
    Module Which Is Tailored For Every Use Case By The Parent Module, Mainly The Number And Names Of Ports.

    See :any:`BaseMod` for arguments, attributes and details.

    Due to the frozen instance approach, implementation specific containers have to be implemented
    via `u.field()`.

    The source code files are typically generated next to the parent module of tailored module.
    Also the module name is based on the parent module and extended by the instance name.
    A :any:`ATailoredMod` requires a Mako template mostly. The default template is typically located
    next to the python file of the tailored module.

    Tailored modules have **two** build methods:

    * `_build`
    * `_build_dep`
    * `_build_final`

    Please take the following points into account:

    * `_build` should be used for all **static** code and initialization.
    * Tailored modules should provide `add` methods to tailor the module instance to the needs of
      the parent module.
    * `_build_dep` is called **after** all `add` methods have been called and should be used for all
      **dynamic** aspects which can only be known after the parent module made its adjustments.

    Attributes:
        filelists: Filelists.

    ??? Example "Module Reference Examples"
        Example Definition:

            >>> import logging
            >>> import ucdp as u
            >>> LOGGER = logging.getLogger(__name__)
            >>> class DmxMod(u.ATailoredMod):
            ...     slaves: u.Namespace = u.Field(default_factory=u.Namespace, init=False)
            ...
            ...     def add_slave(self, name, route=None):
            ...         self.slaves[name] = slave = Slave(dmx=self, name=name)
            ...         portname = f"slv_{name}_o"
            ...         self.add_port(u.UintType(16), portname)
            ...         if route:
            ...             self.con(portname, route)
            ...         return slave
            ...
            ...     def _build(self) -> None:
            ...         pass#self.add_port(u.UintType(16), "mst_i")
            ...
            ...     def _build_dep(self):
            ...         if not self.slaves:
            ...             LOGGER.warning("%r: has no APB slaves", self)
            ...
            >>> class Slave(u.NamedObject):
            ...     dmx: u.BaseMod = u.Field(repr=False)

        Example Usage:

            >>> import ucdp as u
            >>> class TopMod(u.AMod):
            ...     def _build(self) -> None:
            ...         dmx = DmxMod(self, 'u_dmx')
            ...         dmx.add_slave('a')
            ...         dmx.add_slave('b')
            >>> top = TopMod()
            >>> top
            <ucdp.modtailored.TopMod(inst='top', libname='ucdp', modname='top')>
            >>> top.get_inst('u_dmx')
            <ucdp.modtailored.DmxMod(inst='top/u_dmx', libname='ucdp', modname='top_dmx')>
            >>> top.get_inst('u_dmx').slaves
            Namespace([Slave(name='a'), Slave(name='b')])
    """

    filelists: ClassVar[ModFileLists] = ()
    """File Lists."""

    _has_build_dep: bool = PrivateField(default=True)
    _has_build_final: bool = PrivateField(default=True)

    @property
    def modname(self) -> str:
        """Module Name."""
        modname = self.basename
        if self.parent:
            return join_names(self.parent.modname, modname)
        return modname

    @property
    def topmodname(self) -> str:
        """Top Module Name."""
        if self.parent:
            return self.parent.topmodname
        return get_topmodname(self)

    @property
    def libpath(self) -> str:
        """Library Path."""
        if self.parent:
            return self.parent.libpath
        return get_libpath(self.__class__)

    @property
    def is_tb(self) -> bool:
        """Determine if module belongs to Testbench or Design."""
        if self.parent:
            return self.parent.is_tb
        return is_tb_from_modname(self.modname)

    @abstractmethod
    def _build(self) -> None:
        """Build."""

    def _build_dep(self):
        """Build Dependent Parts."""

    def _build_final(self):
        """Build Post."""

    def model_post_init(self, __context: Any) -> None:
        """Run Build."""
        if self.parent:
            self.parent.add_inst(self)

        self._build()
        if not self.parent:
            build(self)



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Identifier.

All symbols within a hardware module are identifier and derived from [Ident][ucdp.ident.Ident].
Identifier should be stored in [Idents][ucdp.ident.Idents].
Identifier are itself part of an expression and therefore a child of [Expr][ucdp.expr.Expr].

??? Example "Identifier Examples"
    Basics:

        >>> import ucdp as u
        >>> class ModeType(u.AEnumType):
        ...     keytype: u.AScalarType = u.UintType(2)
        ...     def _build(self) -> None:
        ...         self._add(0, "add")
        ...         self._add(1, "sub")
        ...         self._add(2, "max")
        >>> class MyType(u.AStructType):
        ...     def _build(self) -> None:
        ...         self._add("mode", ModeType())
        ...         self._add("send", u.ArrayType(u.UintType(8), 3))
        ...         self._add("return", u.UintType(4), u.BWD)
        >>> idents = u.Idents([
        ...     u.Ident(u.UintType(8), "vec_a_i"),
        ...     u.Ident(u.UintType(8), "vec_a_o"),
        ...     u.Ident(u.UintType(4), "vec_c_s"),
        ...     u.Ident(MyType(), "my_a_s"),
        ...     u.Ident(u.ArrayType(MyType(), 4), "my_b_s"),
        ... ])

??? Example "Retrieve An Item"
    Retrieve an item:

        >>> idents['vec_a_i']
        Ident(UintType(8), 'vec_a_i')
        >>> idents['my_a_mode_s']
        Ident(ModeType(), 'my_a_mode_s')

??? Example "Simple Iteration"
    Simple iteration:

        >>> for ident in idents:
        ...     ident
        Ident(UintType(8), 'vec_a_i')
        Ident(UintType(8), 'vec_a_o')
        Ident(UintType(4), 'vec_c_s')
        Ident(MyType(), 'my_a_s')
        Ident(ArrayType(MyType(), 4), 'my_b_s')

??? Example "Unrolling Iteration"
    Unrolling iteration:

        >>> for ident in idents.iter():
        ...     ident
        Ident(UintType(8), 'vec_a_i')
        Ident(UintType(8), 'vec_a_o')
        Ident(UintType(4), 'vec_c_s')
        Ident(MyType(), 'my_a_s')
        Ident(ModeType(), 'my_a_mode_s')
        Ident(ArrayType(UintType(8), 3), 'my_a_send_s')
        Ident(UintType(4), 'my_a_return_s')
        Ident(ArrayType(MyType(), 4), 'my_b_s')
        Ident(ArrayType(UintType(4), 4), 'my_b_return_s')
        Ident(ArrayType(ArrayType(UintType(8), 3), 4), 'my_b_send_s')
        Ident(ArrayType(ModeType(), 4), 'my_b_mode_s')

??? Example "Some Features"
    Some features:

        >>> idents['my_b_send_s']
        Ident(ArrayType(ArrayType(UintType(8), 3), 4), 'my_b_send_s')
        >>> 'my_b_send_s' in idents
        True
        >>> 'my_b_send' in idents
        False

"""

import warnings
from collections import deque
from collections.abc import Callable, Iterable, Iterator
from typing import Any, ClassVar

from .casting import Casting
from .consts import PAT_IDENTIFIER
from .doc import Doc
from .expr import ConcatExpr, ConstExpr, Expr, Log2Expr, MaximumExpr, MinimumExpr, Op, SliceOp, SOp, TernaryExpr
from .ifdef import Ifdefs, join_ifdefs
from .namespace import Namespace
from .nameutil import join_names, split_suffix
from .object import Field, Light, NamedObject, PosArgs
from .orientation import DIRECTION_SUFFIXES, AOrientation
from .typearray import ArrayType
from .typebase import BaseType
from .typestruct import BaseStructType, StructItem


class Ident(Expr, NamedObject, Light):
    """Identifier.

    Args:
        type_: Type.
        name: Name.

    Attributes:
        direction: Direction.
        doc: Documentation Container
        ifdef: IFDEF encapsulation

    ??? Example "Ident Examples"
        Attributes:

            >>> import ucdp as u
            >>> ident = Ident(u.UintType(32), 'base_sub_i')
            >>> ident.type_
            UintType(32)
            >>> ident.name
            'base_sub_i'
            >>> ident.direction
            >>> ident.doc
            Doc()

        Calculated Properties:

            >>> ident.basename
            'base_sub'
            >>> ident.suffix
            '_i'
    """

    type_: BaseType
    name: str = Field(pattern=PAT_IDENTIFIER)
    direction: AOrientation | None = Field(default=None, init=False)
    doc: Doc = Doc()
    ifdefs: Ifdefs = ()

    _posargs: ClassVar[PosArgs] = ("type_", "name")

    def __init__(self, type_: BaseType, name: str, **kwargs):
        super().__init__(type_=type_, name=name, **kwargs)  # type: ignore[call-arg]

    @property
    def basename(self):
        """Base Name."""
        return split_suffix(self.name)[0]

    @property
    def suffix(self):
        """Suffix."""
        return split_suffix(self.name)[1]

    @property
    def ifdef(self) -> str | None:
        """Legacy ifdef."""
        warnings.warn(".ifdef is obsolete. Please use .ifdefs", category=DeprecationWarning, stacklevel=2)
        ifdefs = self.ifdefs
        if ifdefs:
            return ifdefs[0]
        return None

    def __str__(self) -> str:
        return self.name

    def __int__(self):
        return int(getattr(self.type_, "default", -1))

    def __iter__(self):
        for _, ident in _iters([self]):
            yield ident

    def iter(self, filter_=None, stop=None, value=None) -> Iterator:
        """Iterate over Hierarchy."""
        for _, ident in _iters([self], filter_=filter_, stop=stop, value=value):
            yield ident

    def leveliter(self, filter_=None, stop=None, value=None) -> Iterator:
        """Iterate over Hierarchy."""
        yield from _iters([self], filter_=filter_, stop=stop, value=value)

    def cast(self, other: "Ident") -> Casting:
        """Cast self=cast(other)."""
        return None

    def _new_structitem(self, structitem: StructItem, **kwargs):
        direction = self.direction and self.direction * structitem.orientation
        if self.suffix:
            if direction is not None and self.suffix in DIRECTION_SUFFIXES:
                suffix = direction and direction.suffix
            else:
                suffix = self.suffix
        else:
            suffix = ""
        basename = join_names(self.basename, structitem.name)
        return self.new(
            type_=structitem.type_,
            name=f"{basename}{suffix}",
            direction=direction,
            doc=structitem.doc,
            ifdefs=join_ifdefs(self.ifdefs, structitem.ifdefs),
            **kwargs,
        )


#     def iterhier(self, filter_=None, stop=None, maxlevel=None, value=None) -> Iterator:
#        """Iterate over Hierarchy."""
#         hier: List[Ident] = []
#         for ident in self.iter(stop=stop, maxlevel=maxlevel, value=value):
#             hier = hier[: ident.level] + [ident]
#             if not filter_ or filter_(ident):
#                 yield tuple(hier)

#     def get(self, name, value=None):
#        """
#         Get Member of Hierarchy.

#         Args:
#             name: Name

#         Keyword Args:
#             value: value
#             dym (bool): Enriched `ValueError` exception.
#        """
#         if name.startswith("_"):
#             name = f"{self.basename}{name}"
#         return get_ident([self], name, value=value, dym=True)

IdentFilter = Callable[[Ident], bool]
IdentStop = Callable[[Ident], bool]


class Idents(Namespace):
    """
    Identifier Dictionary.

    See examples above.
    """

    def iter(self, filter_: IdentFilter | None = None, stop: IdentStop | None = None) -> Iterator[Ident]:
        """Iterate over all Identifier."""
        for ident in self.values():
            yield from ident.iter(filter_=filter_, stop=stop)

    def leveliter(
        self, filter_: IdentFilter | None = None, stop: IdentStop | None = None
    ) -> Iterator[tuple[int, Ident]]:
        """Iterate over all Identifier with Level."""
        for ident in self.values():
            yield from ident.leveliter(filter_=filter_, stop=stop)

    # def iterhier(self, filter_=None, stop=None, maxlevel=None) -> Iterator:
    #    """Iterate over all Identifier and return hierarchy."""
    #     for ident in self.values():
    #         yield from ident.iterhier(filter_=filter_, stop=stop, maxlevel=maxlevel)

    # def findfirst(self, filter_=None, stop=None, maxlevel=None) -> "Ident" | None:
    #    """Iterate Over Identifier And Find First Match."""
    #     for ident in self.iter(filter_=filter_, stop=stop, maxlevel=maxlevel):
    #         return ident
    #     return None

    def __getitem__(self, name):
        return get_ident(self, name)

    def __contains__(self, item: Any) -> bool:
        if isinstance(item, Ident):
            name = item.name
        elif isinstance(item, str):
            name = item
        else:
            return False
        return _get_ident(self.values(), name) is not None


def _iters(idents: Iterable[Ident], filter_=None, stop=None, value=None, level: int = 0) -> Iterator[tuple[int, Ident]]:  # noqa: C901,PLR0912
    # highly optimized!

    for rootident in idents:
        stack = deque([(level, rootident)])
        while True:
            try:
                identlevel, ident = stack.pop()
            except IndexError:
                break

            if stop and stop(ident):
                break

            type_ = ident.type_
            if value is not None:
                if isinstance(type_, BaseStructType):
                    raise ValueError(f"Cannot apply value {value} for {type_}")
                type_ = type_.new(default=value)
                ident = ident.new(type_=type_)

            if not filter_ or filter_(ident):
                yield identlevel, ident
            if isinstance(type_, BaseStructType):
                substack = deque()
                for structitem in type_.values():
                    child = ident._new_structitem(structitem)
                    if not stop or not stop(child):
                        substack.append((identlevel + 1, child))
                stack.extend(reversed(substack))
            elif isinstance(type_, ArrayType):
                elementident = ident.new(type_=type_.itemtype)
                # do not forward stop and filter_ as these are just intermediate identifiers
                subidents = tuple(_iters([elementident], level=identlevel + 1))
                for sublevel, subident in subidents[1:]:
                    # create array with new element type
                    childtype = type_.new(itemtype=subident.type_)
                    child = subident.new(type_=childtype)
                    if not stop or not stop(child):
                        stack.append((sublevel, child))


def get_ident(idents: Iterable[Ident], name: str, value=None, dym=False) -> Ident | None:
    """Retrieve identifier by `name`, without iterating over the entire identifier tree."""
    assert value is None, "TODO"
    ident = _get_ident(idents, name)
    if ident is not None:
        return ident

    # Beautiful error handling
    msg = f"'{name}' is not known."
    #     if dym and any(idents):
    #         names = [ident.name for ident in _iters(idents)]
    #         nametree = align(_get_identtree(idents), rtrim=True)
    #         msg += f" Known are\n{nametree}\n\n{msg}\n"
    #         msg += didyoumean(name, names, multiline=True)
    raise ValueError(msg)


def get_subname(parent: Ident, ident: Ident):
    """Get name relative to parent."""
    if parent is ident:
        return ident.suffix
    parentbasename = parent.basename
    return ident.name.removeprefix(parentbasename)[1:]


# def get_subnames(idents):
#    """Return names of hierarchical idents."""
#     names = []
#     prefix = ""
#     for ident in idents:
#         basename = ident.basename
#         names.append(basename.removeprefix(prefix))
#         prefix = f"{basename}_"
#     return names


def _get_ident(idents, name) -> Ident | None:
    # TODO: optimize
    for _, ident in _iters(idents):
        if ident.name == name:
            return ident
    return None


# def _get_identtree(idents, pre=""):
#     for ident in _iters(idents):
#         pre = "  " * ident.level
#         yield f"{pre}'{ident.name}'", ident.type_


def get_expridents(expr: Expr) -> tuple[Ident, ...]:
    """
    Determine used identifier in `expr`.
    """
    idents = {}
    heap = deque((expr,))
    while heap:
        item = heap.popleft()
        if isinstance(item, ConstExpr):
            pass
        elif isinstance(item, Op):
            heap.append(item.left)
            heap.append(item.right)
        elif isinstance(item, (SOp, SliceOp)):
            heap.append(item.one)
        elif isinstance(item, (ConcatExpr, MinimumExpr, MaximumExpr)):
            heap.extend(item.items)
        elif isinstance(item, TernaryExpr):
            heap.append(item.cond.left)
            heap.append(item.cond.right)
            heap.append(item.one)
            heap.append(item.other)
        elif isinstance(item, Log2Expr):
            heap.append(item.expr)
        elif isinstance(item, Ident):
            if item.name not in idents:
                idents[item.name] = item
        else:
            raise TypeError(f"Unknown expr {item}")  # pragma: no cover
    return tuple(idents.values())



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Command Line Interface."""

import importlib
import os
from functools import lru_cache
from importlib.metadata import entry_points
from pathlib import Path

import click

from .logging import LOGGER
from .util import extend_sys_path

Paths = tuple[Path, ...]
PATHS = tuple(Path(path) for path in os.environ.get("UCDP_PATH", "").split())


@lru_cache
def _find_commands(paths: Paths):
    return dict(__find_commands(paths))


def __find_commands(paths: Paths):  # pragma: no cover
    """Find Commands."""
    for entry_point in entry_points(group="ucdp.cli"):
        yield entry_point.name, entry_point.value
    if paths:
        with extend_sys_path(paths, use_env_default=True):
            for path in paths:
                for clifile in path.glob("*/cli.py"):
                    libname, modname = clifile.parts[-2], clifile.stem
                    climod = importlib.import_module(f"{libname}.cli")
                    clicmds = getattr(climod, "UCDP_COMMANDS", [])
                    if not isinstance(clicmds, (list, tuple)):
                        LOGGER.warning("Invalid UCDP_COMMANDS in %s", clifile)
                        continue
                    for clicmd in clicmds:
                        yield clicmd, f"{libname}.{modname}.{clicmd}"


class MainGroup(click.Group):  # pragma: no cover
    """Main Command Group with Dynamically Loaded Commands."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def list_commands(self, ctx):
        """List Commands."""
        base = super().list_commands(ctx)
        lazy = list(_find_commands(PATHS))
        return sorted(base + lazy)

    def get_command(self, ctx, name):
        """Load Command."""
        if name in _find_commands(PATHS):
            try:
                return self._load(name)
            except Exception as exc:  # pragma: no cover
                LOGGER.error("Could not load command '%s' (%s)", name, exc)
        return super().get_command(ctx, name)

    def _load(self, name):
        import_path = _find_commands(PATHS)[name]
        modname, cmd_object_name = import_path.rsplit(".", 1)
        mod = importlib.import_module(modname)
        return getattr(mod, cmd_object_name)



# MIT License
#
# Copyright (c) 2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Create Utilities."""

from pathlib import Path
from typing import Literal

from caseconverter import pascalcase, snakecase, titlecase
from makolator import Datamodel

from .consts import PAT_IDENTIFIER_LOWER
from .generate import get_makolator
from .object import Field, Object

TYPE_CHOICES = [
    "AConfigurableMod",
    "AConfigurableTbMod",
    "AGenericTbMod",
    "AMod",
    "ATailoredMod",
    "ATbMod",
]

TYPE_CHOICES_TB = [
    "AConfigurableTbMod",
    "AGenericTbMod",
    "ATbMod",
]

TB_MAP = {
    "AConfigurableMod": "AGenericTbMod",
    "AMod": "ATbMod",
    "ATailoredMod": "AGenericTbMod",
}


class CreateInfo(Object):
    """Module Skeleton Information."""

    module: str = Field(pattern=PAT_IDENTIFIER_LOWER)
    """Module Name."""

    library: str = Field(pattern=PAT_IDENTIFIER_LOWER)
    """Library Name."""

    regf: bool = False
    """Include Register File."""

    descr: str = ""
    """Description."""

    flavour: Literal["AConfigurableMod", "AConfigurableTbMod", "AGenericTbMod", "AMod", "ATailoredMod", "ATbMod"] = (
        "AMod"
    )
    """Flavour."""

    @property
    def module_pascalcase(self) -> str:
        """Module Name In Pascalcase."""
        return pascalcase(self.module)

    @property
    def module_snakecase(self) -> str:
        """Module Name In Snakecase."""
        return snakecase(self.module)

    @property
    def module_titlecase(self) -> str:
        """Module Name In Titlecase."""
        return titlecase(self.module)

    @property
    def descr_or_default(self) -> str:
        """Module Description Or A Nice Default."""
        if self.descr:
            return self.descr
        return f"{self.module_titlecase} Module"

    @property
    def is_tb(self) -> bool:
        """Module Name In Titlecase."""
        return self.flavour in TYPE_CHOICES_TB


def create(info: CreateInfo, force: bool) -> None:
    """Creates A Module Skeleton Based On `info`."""
    mklt = get_makolator(force=force)
    mklt.datamodel = Datamodel(info=info)
    mklt.gen([Path("mod.py.mako")], dest=Path(f"{info.library}/{info.module}.py"))



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Dictionary.
"""

from typing import Any

from .object import Object, PrivateField


class Dict(Object):
    """
    Re-Implementation of python dictorary, compatible with [Object][ucdp.object.Object].

    ??? Example "Dictorary Examples"
        Basics:

            >>> import ucdp as u
            >>> data = u.Dict()
            >>> data
            Dict()

        Add items:

            >>> data['a'] = 1
            >>> data['c'] = 3
            >>> data['b'] = 2

        Get items:

            >>> data['a']
            1
            >>> data['d']
            Traceback (most recent call last):
            ...
            KeyError: 'd'

        Iteration:

            >>> tuple(data)
            ('a', 'c', 'b')

        Keys:

            >>> data.keys()
            dict_keys(['a', 'c', 'b'])

        Values:

            >>> data.values()
            dict_values([1, 3, 2])

        Key-Value Pairs:

            >>> data.items()
            dict_items([('a', 1), ('c', 3), ('b', 2)])
    """

    _items: dict[Any, Any] = PrivateField(default_factory=dict)

    def __iter__(self):
        yield from self._items.keys()

    def keys(self):
        """Return Dictionary Keys."""
        return self._items.keys()

    def values(self):
        """Return Dictionary Value."""
        return self._items.values()

    def items(self):
        """Return Key-Value Pairs."""
        return self._items.items()

    def get(self, *args, **kwargs):
        """Retrieve Element."""
        return self._items.get(*args, **kwargs)

    def __getitem__(self, key):
        return self._items[key]

    def __setitem__(self, key: Any, value: Any):
        self._items[key] = value

    def __len__(self) -> int:
        return len(self._items)

    def __contains__(self, key) -> bool:
        return key in self._items



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Configuration."""

import pydantic

import ucdp as u


def test_example_simple(example_simple):
    """Example."""
    from glbl_lib.clk_gate import ClkGateMod
    from uart_lib.uart import UartMod

    assert tuple(u.get_baseclassinfos(UartMod)) == (
        u.BaseClassInfo(cls=UartMod, libname="uart_lib", modname="uart", clsname="UartMod"),
        u.BaseClassInfo(cls=u.AMod, libname="ucdp", modname="mod", clsname="AMod"),
        u.BaseClassInfo(cls=u.BaseTopMod, libname="ucdp", modname="modbasetop", clsname="BaseTopMod"),
        u.BaseClassInfo(cls=u.BaseMod, libname="ucdp", modname="modbase", clsname="BaseMod"),
        u.BaseClassInfo(cls=u.NamedObject, libname="ucdp", modname="object", clsname="NamedObject"),
        u.BaseClassInfo(cls=u.Object, libname="ucdp", modname="object", clsname="Object"),
        u.BaseClassInfo(cls=pydantic.main.BaseModel, libname="pydantic", modname="main", clsname="BaseModel"),
    )

    assert tuple(u.get_baseclassinfos(UartMod())) == tuple(u.get_baseclassinfos(UartMod))

    assert tuple(u.get_baseclassinfos(ClkGateMod)) == (
        u.BaseClassInfo(cls=ClkGateMod, libname="glbl_lib", modname="clk_gate", clsname="ClkGateMod"),
        u.BaseClassInfo(cls=u.AMod, libname="ucdp", modname="mod", clsname="AMod"),
        u.BaseClassInfo(cls=u.BaseTopMod, libname="ucdp", modname="modbasetop", clsname="BaseTopMod"),
        u.BaseClassInfo(cls=u.BaseMod, libname="ucdp", modname="modbase", clsname="BaseMod"),
        u.BaseClassInfo(cls=u.NamedObject, libname="ucdp", modname="object", clsname="NamedObject"),
        u.BaseClassInfo(cls=u.Object, libname="ucdp", modname="object", clsname="Object"),
        u.BaseClassInfo(cls=pydantic.main.BaseModel, libname="pydantic", modname="main", clsname="BaseModel"),
    )



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


"""
Assignment Handling.

The class [Assigns][ucdp.assigns.Assigns] manages sets of signal assignments.
Either statically in modules but also within flip-flop and multiplexer definitions.

??? Example "Basic Examples"
    All of the following is happening within any hardware module, flip-flop or multiplexer,
    but can also be used within own code.

        >>> import ucdp as u
        >>> signals = u.Idents([
        ...     u.Port(u.ClkRstAnType(), "main_i"),
        ...     u.Port(u.UintType(8), "vec_a_i"),
        ...     u.Port(u.UintType(8), "vec_a_o"),
        ...     u.Port(u.UintType(14), "vec_b_i"),
        ...     u.Port(u.UintType(14), "vec_b_o"),
        ...     u.Port(u.UintType(14), "vec_c_o"),
        ...     u.Signal(u.ClkRstAnType(), "main_s"),
        ...     u.Signal(u.UintType(8), "vec_a_s"),
        ...     u.Signal(u.UintType(4), "vec_b_s"),
        ...     u.Signal(u.UintType(4), "vec_c_s"),
        ... ])
        >>> assigns = u.Assigns(targets=signals, sources=signals)
        >>> assigns.set_default(signals['vec_a_o'], signals['vec_a_i'])
        >>> assigns.set_default(signals['vec_b_o'], signals['vec_b_i'])
        >>> assigns.set(signals['vec_a_o'], signals['vec_a_s'])
        >>> for assign in assigns:
        ...     str(assign)
        'vec_a_o  <----  vec_a_s'
        'vec_b_o  <----  vec_b_i'

??? failure "Multiple Assignments"
    Multiple assignments are forbidden:

        >>> assigns.set(signals['vec_a_o'], signals['vec_a_s'])
        Traceback (most recent call last):
        ...
        ValueError: 'vec_a_o' already assigned to 'vec_a_s'

??? Example "Default Examples"
    Defaults are managed separately:

        >>> for assign in assigns.defaults():
        ...     str(assign)
        'vec_a_o  <----  vec_a_i'
        'vec_b_o  <----  vec_b_i'

??? Example "Mapping"
    With `inst=True` the all target signals are mapped:

        >>> assigns = u.Assigns(targets=signals, sources=signals, inst=True)
        >>> assigns.set_default(signals['vec_a_i'], signals['vec_a_i'])
        >>> assigns.set_default(signals['vec_b_i'], signals['vec_b_i'])
        >>> assigns.set(signals['vec_a_i'], signals['vec_a_s'])
        >>> for assign in assigns:
        ...     str(assign)
        'main_i  ---->  None'
        'main_clk_i  ---->  None'
        'main_rst_an_i  ---->  None'
        'vec_a_i  ---->  vec_a_s'
        'vec_a_o  <----  None'
        'vec_b_i  ---->  vec_b_i'
        'vec_b_o  <----  None'
        'vec_c_o  <----  None'
        'main_s  ---->  None'
        'main_clk_s  ---->  None'
        'main_rst_an_s  ---->  None'
        'vec_a_s  ---->  None'
        'vec_b_s  ---->  None'
        'vec_c_s  ---->  None'

"""

from collections.abc import Callable, Iterator
from typing import Any

from ._castingnamespace import CastingNamespace
from ._sliceassign import SliceAssign
from .casting import Casting
from .drivers import Drivers, Source, Target
from .exceptions import DirectionError, LockError
from .expr import Expr, SliceOp
from .exprparser import ExprParser
from .ident import Ident, Idents, get_subname
from .nameutil import split_suffix
from .note import Note
from .object import Object, PrivateField, model_validator
from .orientation import BWD, FWD, IN, INOUT, OUT, Direction
from .signal import BaseSignal, Port

_DIRECTION_MAP = {
    IN: "---->",
    OUT: "<----",
    INOUT: "<--->",
    None: None,
}

TargetFilter = Callable[[BaseSignal], bool]


class Assign(Object):
    """
    A Single Assignment of `expr` to `target`.

    Attributes:
        target: Assigned identifier.
        source: Assigned expression.
    """

    target: BaseSignal
    source: Source | None = None

    @property
    def name(self) -> str | None:
        """Name."""
        return self.target.name

    @property
    def type_(self):
        """Type."""
        return self.target.type_

    @property
    def doc(self):
        """Doc."""
        return self.target.doc

    @property
    def direction(self) -> Direction:
        """Direction."""
        return Direction.cast(self.target.direction)  # type: ignore[return-value]

    @property
    def ifdef(self) -> str | None:
        """IFDEF."""
        return self.target.ifdef

    @property
    def ifdefs(self) -> str | None:
        """IFDEF."""
        return self.target.ifdefs

    @property
    def sign(self) -> str | None:
        """Sign."""
        return _DIRECTION_MAP[self.direction]

    def __str__(self):
        return f"{self.target}  {self.sign}  {self.source}"


_TargetAssigns = dict[str, Source | SliceAssign | None]


class Assigns(Object):
    """
    Assignments.

    An instance of [Assigns][ucdp.assigns.Assigns] manages a set of signal assignments.

    Attributes:
        targets: Identifiers allowed to be assigned.
        source: Identifiers allowed to be used in assignment. `targets` by default.
        drivers: Driver tracking, to avoid multiple drivers. To be shared between multiple assignments,
                        where only one driver is allowed.
        inst: All Instances Assignment Mode.
        sub: Sublevel Instance Assignment Mode.

    """

    targets: Idents
    sources: Idents
    drivers: Drivers | None = None
    inst: bool = False
    _defaults: _TargetAssigns = PrivateField(default_factory=dict)
    _assigns: _TargetAssigns = PrivateField(default_factory=dict)
    __is_locked: bool = PrivateField(default=False)

    @property
    def is_locked(self) -> bool:
        """Locked."""
        return self.__is_locked

    def lock(self) -> None:
        """Lock."""
        if self.__is_locked:
            raise LockError("Assigns are already locked. Cannot lock again.")
        self.__is_locked = True

    @model_validator(mode="before")
    @classmethod
    def __pre_init(cls, data: Any) -> Any:
        if isinstance(data, dict):
            data["sources"] = data.get("sources") or data.get("targets")
        return data

    def set_default(
        self,
        target: Target,
        source: Source,
        cast: bool | None = False,
        overwrite: bool = False,
        filter_: TargetFilter | None = None,
    ):
        """Set Default of `target` to `source`.

        Params:
            target: Target.
            source: Source.
            cast: cast to target.
            overwrite: overwrite target.
        """
        if self.__is_locked:
            raise LockError(f"Cannot set default '{source}' to '{target}'")
        casting = self._prepare(target, source, cast)
        self._set(
            self._defaults,
            target,
            source,
            casting,
            overwrite=overwrite,
            filter_=filter_,
        )

    def set(
        self,
        target: BaseSignal,
        source: Source,
        cast: bool | None = False,
        overwrite: bool = False,
        filter_: TargetFilter | None = None,
    ):
        """Set Assignment of `target` to `source`.

        Params:
            target: Target.
            source: Source.
            cast: cast to target.
            overwrite: overwrite target.
        """
        if self.__is_locked:
            raise LockError(f"Cannot set '{source}' to '{target}'")
        casting = self._prepare(target, source, cast)
        self._set(
            self._assigns,
            target,
            source,
            casting,
            drivers=self.drivers,
            overwrite=overwrite,
            filter_=filter_,
        )

    def get(self, target: BaseSignal) -> Source | None:
        """Get Assignment of `target`."""
        source = self._get(self._assigns, target)
        if source is None:
            source = self._get(self._defaults, target)
        return source

    def _prepare(self, target: Target, source: Source, cast: bool | None) -> Casting | None:
        # Note
        if isinstance(source, Note):
            return None

        if not isinstance(source, (Expr, Note)):
            raise ValueError(f"Source {source} is not a Expression or Note")
        if not isinstance(target, (BaseSignal, SliceOp)):
            raise ValueError(f"Target {target} is not a Signal, Port or Slice of them")

        # Normalize Directions
        orient = BWD if self.inst else FWD
        targetdir = isinstance(target, Port) and target.direction and (target.direction * orient)
        sourcedir = isinstance(source, Port) and source.direction and source.direction

        # Check Direction
        if targetdir == sourcedir == IN:
            raise DirectionError(f"Cannot drive '{target}' by '{source}'")

        # Check casting
        is_connectable = target.type_.is_connectable(source.type_) or source.type_.is_connectable(target.type_)
        if cast is False:
            # casting is forbidden
            if not is_connectable:
                hint = self._get_hint(target, source)
                msg = f"Cannot assign '{source}' of type {source.type_} to '{target}' of type {target.type_}.{hint}"
                raise TypeError(msg)
            casting = None
        else:
            casting = self._get_casting(target, source)
            if casting is None:
                if cast is True or not is_connectable:
                    # casting required
                    hint = self._get_hint(target, source)
                    msg = f"Cannot cast '{source}' of {source.type_} to '{target}' of {target.type_}.{hint}"
                    raise TypeError(msg)
        return casting

    @staticmethod
    def _get_casting(target: Expr | Note, source: Expr | Note) -> Casting:
        if isinstance(source, Ident) and isinstance(target, Ident):
            return target.cast(source)
        return None

    def _get_hint(self, target: Target, source: Source) -> str:
        if self._get_casting(target, source) is not None:
            return " Try to cast."

        if self._get_casting(source, target) is not None:
            return " Try to cast, but swap target and source."

        if (
            isinstance(target, Expr)
            and isinstance(source, Expr)
            and (target.type_.is_connectable(source.type_) or source.type_.is_connectable(target.type_))
        ):
            return " You do NOT need to cast!"

        return ""

    def _set(  # noqa: C901, PLR0912
        self,
        assigns: _TargetAssigns,
        target: Target,
        source: Source,
        casting: Casting | None,
        drivers: Drivers | None = None,
        overwrite: bool = False,
        filter_: TargetFilter | None = None,
    ):
        if isinstance(source, Note):
            if isinstance(target, SliceOp):
                raise ValueError(f"Cannot set {source} on slice {target}")
            for subtarget in target.iter(filter_=filter_):
                assigns[subtarget.name] = source
        elif isinstance(target, SliceOp):
            self._setslice(assigns, drivers, target, source, overwrite, filter_)
        else:
            # note: expression-assignments are forbidden on inst=True
            # note: check for valid identifiers?
            if casting:
                subtargets, subsources = self._cast(target, source, casting)  # type: ignore[arg-type]
            else:
                is_sourceident = isinstance(source, Ident)
                subtargets = tuple(target.iter())
                if len(subtargets) != 1 and not is_sourceident:
                    raise ValueError("TODO-x")
                subsources = tuple(source.iter()) if is_sourceident else [source]  # type: ignore[attr-defined,assignment]
            for subtarget, subsource in zip(subtargets, subsources, strict=True):
                if filter_ and not filter_(subtarget):
                    continue

                # reverse
                if not self.inst and not subtarget.direction != IN:
                    subtarget, subsource = subsource, subtarget  # noqa: PLW2901

                if not isinstance(subtarget, BaseSignal):
                    raise ValueError(f"Target {subtarget} is not a Signal or Port")

                if not isinstance(subsource, Expr):
                    raise ValueError(f"Target {subsource} is not an Expression")

                # check
                if subtarget.name in assigns:
                    if not overwrite:
                        raise ValueError(f"'{subtarget}' already assigned to '{assigns[subtarget.name]}'")

                # drivers
                if drivers is not None:
                    drivers.set(subtarget, subsource, overwrite=overwrite)

                # assign
                assigns[subtarget.name] = subsource

    @staticmethod
    def _cast(target: Ident, source: Ident, targetcasting):
        """Cast `source` to `target` with `targetcasting`."""
        subsources = []
        subtargets = []
        casting = dict(targetcasting)
        parser = ExprParser(namespace=CastingNamespace(source=source))
        for subtarget in target.iter():
            subtargetname = get_subname(target, subtarget)
            subtargetbasename = split_suffix(subtargetname)[0]
            subcast = casting.pop(subtargetbasename, None)
            subsource: Expr
            if subcast is None:
                continue
            if subcast == "":
                subsource = source
            else:
                try:
                    subsource = parser(subcast)
                except NameError as err:
                    raise NameError(f"{err} in casting {subcast!r} to '{subtarget}'") from None
            subsources.append(subsource)
            subtargets.append(subtarget)
        return subtargets, subsources

    def __iter__(self):
        return self.iter()

    def iter(self, filter_=None) -> Iterator[Assign]:
        """Iterate over assignments."""
        defaults = self._defaults
        assigns = self._assigns
        for ident in self.targets.iter():
            target: BaseSignal = ident  # type: ignore[assignment]
            if filter_ and not filter_(target):
                continue
            default = self._get(defaults, target)
            source = self._get(assigns, target, default=default)
            if source is not None or self.inst:
                yield Assign(target=target, source=source)

    def defaults(self) -> Iterator[Assign]:
        """Iterate Over Defaults."""
        defaults = self._defaults
        assigns = self._assigns
        for ident in self.targets.iter():
            target: BaseSignal = ident  # type: ignore[assignment]
            default = self._get(defaults, target)
            if self.inst or default is not None or bool(assigns.get(target.name, None)):
                yield Assign(target=target, source=default)

    @staticmethod
    def _get(assigns: _TargetAssigns, target: BaseSignal, default: Source | None = None) -> Source | None:
        source = assigns.get(target.name, default)
        if isinstance(source, SliceAssign):
            return source.get_expr(default)
        return source

    @staticmethod
    def _setslice(
        assigns: _TargetAssigns,
        drivers: Drivers | None,
        target: SliceOp,
        source: Expr,
        overwrite: bool,
        filter_: TargetFilter | None,
    ):
        if not isinstance(target.one, BaseSignal):
            raise ValueError(f"Slice target {target.one} is not a Signal or Port")
        targetident: BaseSignal = target.one
        if filter_ and not filter_(targetident):
            return
        try:
            sliceassign = assigns[targetident.name]
        except KeyError:
            sliceassign = None

        if sliceassign is not None and not isinstance(sliceassign, SliceAssign) and not overwrite:
            raise ValueError(f"'{targetident}' already assigned to '{assigns[targetident.name]}'")

        if sliceassign is None or not isinstance(sliceassign, SliceAssign):
            sliceassign = assigns[targetident.name] = SliceAssign(type_=targetident.type_, expr=targetident)
            if drivers is not None:
                drivers.set(targetident, sliceassign, overwrite=overwrite)

        sliceassign.set(target.slice_, source, overwrite)



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Utilities."""

import sys
from inspect import getfile
from pathlib import Path

import ucdp as u

COPYRIGHT = [
    "",
    " MIT License",
    "",
    " Copyright (c) 2024-2025 nbiotcloud",
    "",
    " Permission is hereby granted, free of charge, to any person obtaining a copy",
    ' of this software and associated documentation files (the "Software"), to deal',
    " in the Software without restriction, including without limitation the rights",
    " to use, copy, modify, merge, publish, distribute, sublicense, and/or sell",
    " copies of the Software, and to permit persons to whom the Software is",
    " furnished to do so, subject to the following conditions:",
    "",
    " The above copyright notice and this permission notice shall be included in all",
    " copies or substantial portions of the Software.",
    "",
    ' THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR',
    " IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,",
    " FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE",
    " AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER",
    " LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,",
    " OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE",
    " SOFTWARE.",
    "",
]


def test_get_copyright(example_simple):
    """Test Copyright."""
    top = u.load("glbl_lib.clk_gate", paths=None)
    assert u.get_copyright(top.mod).splitlines() == COPYRIGHT
    assert u.get_copyright(Path(getfile(top.mod.__class__))).splitlines() == COPYRIGHT


def test_extend_sys_path(tmp_path):
    """Test Extend Sys Path."""
    before = tuple(sys.path)

    with u.extend_sys_path([]):
        pass
    assert before == tuple(sys.path)

    assert str(tmp_path) not in sys.path
    with u.extend_sys_path([tmp_path]):
        assert str(tmp_path) in sys.path

    assert before == tuple(sys.path)



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Command-Line-Interface."""

import subprocess
from pathlib import Path

from click.testing import CliRunner
from contextlib_chdir import chdir
from pydantic import BaseModel
from pytest import mark
from test2ref import assert_refdata, configure

import ucdp as u

from .conftest import TESTDATA_PATH

configure(ignore_spaces=True)

REPLACEMENTS = ((Path("$PRJROOT"), "$PRJROOT"),)


class Result(BaseModel):
    """Result."""

    exit_code: int
    stdout: str
    stderr: str


def run(*cmd, exit_code: int = 0) -> Result:
    """Run."""
    result = subprocess.run(("ucdp", *cmd), check=False, capture_output=True, text=True)  # noqa: S603
    assert result.returncode == exit_code
    stdout = result.stdout
    stderr = result.stderr
    return Result(exit_code=result.returncode, stdout=stdout, stderr=stderr)


def run2console(prjroot: Path, *cmd, exit_code: int = 0, output="console.txt") -> None:
    """Run And Capture Console."""
    result = run(*cmd, exit_code=exit_code)
    if result.stdout:
        (prjroot / output).write_text(result.stdout)
    if result.stderr:
        (prjroot / output).with_suffix(".err.txt").write_text(result.stderr)


def test_check(prjroot, example_simple):
    """Check Command."""
    run2console(prjroot, "check", "uart_lib.uart", output="check-uart.txt")

    run("check", "uart_lib.uart2", exit_code=1)

    run2console(prjroot, "check", "uart_lib.uart", "--stat", output="check-stat.txt")

    assert_refdata(test_check, prjroot)


def test_gen(prjroot, example_simple, uartcorefile):
    """Generate and Clean Command."""
    uartfile = prjroot / "uart_lib" / "uart" / "rtl" / "uart.sv"

    assert not uartfile.exists()

    run2console(prjroot, "gen", "uart_lib.uart", "-f", "hdl", output="gen.txt")

    assert uartfile.exists()

    run2console(prjroot, "cleangen", "uart_lib.uart", "-f", "hdl", output="cleangen.txt")

    assert not uartfile.exists()

    assert_refdata(test_gen, prjroot)


def test_gen_check(prjroot, example_simple, uartcorefile):
    """Generate with Check."""
    uartfile = prjroot / "uart_lib" / "uart" / "rtl" / "uart.sv"

    run2console(prjroot, "gen", "uart_lib.uart", "--check", exit_code=1, output="check.txt")

    assert uartfile.exists()

    run2console(prjroot, "gen", "uart_lib.uart", "--check", output="re-check.txt")

    assert_refdata(test_gen_check, prjroot)


def test_gen_tb(prjroot, example_simple):
    """Generate with Check."""
    run2console(prjroot, "gen", "glbl_lib.regf_tb#uart_lib.uart-uart_lib.uart_regf", output="gen.txt")
    assert_refdata(test_gen_tb, prjroot)


def test_gen_topsfile(prjroot, example_simple, uartcorefile):
    """Generate with --tops-file."""
    topfile = prjroot / "tops.txt"
    topfile.write_text("""
# comment
  uart_lib.uart

""")

    run2console(prjroot, "gen", "--tops-file", str(topfile))

    assert_refdata(test_gen_topsfile, prjroot)


def test_gen_default(prjroot, example_simple, uartcorefile):
    """Generate and Clean Command."""
    run2console(prjroot, "gen", "uart_lib.uart", output="gen.txt")

    run2console(prjroot, "cleangen", "uart_lib.uart", output="cleangen.txt")

    assert_refdata(test_gen_default, prjroot)


def test_filelist(prjroot, example_simple):
    """Filelist Command."""
    run2console(prjroot, "filelist", "uart_lib.uart", "-f", "hdl")
    assert_refdata(test_filelist, prjroot, replacements=REPLACEMENTS)


def test_filelist_default(prjroot, example_simple):
    """Filelist Command with Default."""
    run2console(prjroot, "filelist", "uart_lib.uart")
    assert_refdata(test_filelist_default, prjroot, replacements=REPLACEMENTS)


def test_filelist_file(prjroot, example_simple):
    """Filelist Command."""
    filepath = prjroot / "file.txt"
    run2console(prjroot, "filelist", "uart_lib.uart", "-f", "hdl", "--file", str(filepath))
    assert_refdata(test_filelist_file, prjroot, replacements=REPLACEMENTS)


def test_filelist_other(prjroot, example_filelist):
    """Filelist Command."""
    run2console(prjroot, "filelist", "filelist_lib.filelist", "-f", "hdl")
    assert_refdata(test_filelist_other, prjroot, replacements=REPLACEMENTS)


def test_fileinfo(prjroot, example_simple):
    """Fileinfo Command."""
    run2console(prjroot, "fileinfo", "uart_lib.uart", "-f", "hdl")
    assert_refdata(test_fileinfo, prjroot, replacements=REPLACEMENTS)


def test_fileinfo_default(prjroot, example_simple):
    """Fileinfo Command."""
    run2console(prjroot, "fileinfo", "uart_lib.uart")
    assert_refdata(test_fileinfo_default, prjroot, replacements=REPLACEMENTS)


def test_fileinfo_minimal(prjroot, example_simple):
    """Fileinfo Command Minimal."""
    run2console(prjroot, "fileinfo", "uart_lib.uart", "-m")
    assert_refdata(test_fileinfo_minimal, prjroot, replacements=REPLACEMENTS)


def test_fileinfo_maxlevel(prjroot, example_simple):
    """Fileinfo Command with Maxlevel."""
    run2console(prjroot, "fileinfo", "uart_lib.uart", "-f", "hdl", "--maxlevel=1")
    assert_refdata(test_fileinfo_maxlevel, prjroot, replacements=REPLACEMENTS)


def test_fileinfo_file(prjroot, example_simple):
    """Fileinfo Command with File."""
    filepath = prjroot / "file.txt"
    run2console(prjroot, "fileinfo", "uart_lib.uart", "-f", "hdl", "--file", str(filepath))
    assert_refdata(test_fileinfo_file, prjroot, replacements=REPLACEMENTS)


def test_overview(prjroot, example_simple):
    """Overview Command."""
    run2console(prjroot, "overview", "uart_lib.uart")
    assert_refdata(test_overview, prjroot)


def test_overview_minimal(prjroot, example_simple):
    """Overview Command - Minimal."""
    run2console(prjroot, "overview", "uart_lib.uart", "-m")
    assert_refdata(test_overview_minimal, prjroot)


def test_overview_file(prjroot, example_simple):
    """Overview Command - Minimal."""
    filepath = prjroot / "file.txt"
    run2console(prjroot, "overview", "uart_lib.uart", "-o", str(filepath))
    assert_refdata(test_overview_file, prjroot)


def test_overview_tags(prjroot, example_simple):
    """Overview Command."""
    run2console(prjroot, "overview", "uart_lib.uart", "--tag", "intf")
    assert_refdata(test_overview_tags, prjroot)


def test_info_examples(prjroot, example_simple):
    """Info Examples Command."""
    run2console(prjroot, "info", "examples")
    assert_refdata(test_info_examples, prjroot)


def test_info_templates(prjroot, example_simple):
    """Info Templates Command."""
    run2console(prjroot, "info", "template-paths")
    assert_refdata(test_info_templates, prjroot)


def test_rendergen(prjroot, example_simple):
    """Command rendergen."""
    template_filepath = TESTDATA_PATH / "example.txt.mako"
    filepath = prjroot / "output.txt"
    run2console(prjroot, "rendergen", "uart_lib.uart", str(template_filepath), str(filepath))
    assert_refdata(test_rendergen, prjroot)


def test_rendergen_defines(prjroot, example_simple):
    """Command rendergen."""
    template_filepath = TESTDATA_PATH / "example.txt.mako"
    filepath = prjroot / "output.txt"
    run2console(
        prjroot, "rendergen", "uart_lib.uart", str(template_filepath), str(filepath), "-D", "one=1", "-D", "two"
    )
    assert_refdata(test_rendergen_defines, prjroot)


def test_renderinplace(prjroot, example_simple):
    """Command renderinplace."""
    template_filepath = TESTDATA_PATH / "example.txt.mako"
    filepath = prjroot / "output.txt"
    filepath.write_text("""
GENERATE INPLACE BEGIN content('test')
GENERATE INPLACE END content
""")
    run2console(prjroot, "renderinplace", "uart_lib.uart", str(template_filepath), str(filepath))
    assert_refdata(test_renderinplace, prjroot)


def test_ls(prjroot, example_simple):
    """List Command."""
    run2console(prjroot, "ls")
    assert_refdata(test_ls, prjroot)


def test_ls_base(prjroot, example_simple):
    """List Command."""
    run2console(prjroot, "ls", "-A")
    assert_refdata(test_ls_base, prjroot)


def test_ls_local(prjroot, example_simple):
    """List Command - Local Only."""
    run2console(prjroot, "ls", "--local")
    assert_refdata(test_ls_local, prjroot)


def test_ls_nonlocal(prjroot, example_simple):
    """List Command - Non-Local Only."""
    run2console(prjroot, "ls", "--no-local", exit_code=1)
    assert_refdata(test_ls_nonlocal, prjroot)


def test_ls_filepath(prjroot, tests):
    """List Command With Filepath."""
    run2console(prjroot, "ls", "-fn")
    assert_refdata(test_ls_filepath, prjroot, replacements=((Path("tests"), "tests"),))


def test_ls_filepath_abs(prjroot, tests):
    """List Command With Filepath."""
    run2console(prjroot, "ls", "-Fn")
    assert_refdata(test_ls_filepath_abs, prjroot)


def test_ls_names(prjroot, tests):
    """List with Names Only."""
    run2console(prjroot, "ls", "-n")
    assert_refdata(test_ls_names, prjroot)


def test_ls_top(prjroot, tests):
    """List Top Modules Only."""
    run2console(prjroot, "ls", "-t")
    assert_refdata(test_ls_top, prjroot)


def test_ls_notop(prjroot, tests):
    """List No Top Modules Only."""
    run2console(prjroot, "ls", "-T")
    assert_refdata(test_ls_notop, prjroot)


def test_ls_tb(prjroot, tests):
    """List Testbenches Only."""
    run2console(prjroot, "ls", "-b")
    assert_refdata(test_ls_tb, prjroot)


def test_ls_notb(prjroot, tests):
    """List Testbenches Only."""
    run2console(prjroot, "ls", "-B")
    assert_refdata(test_ls_notb, prjroot)


def test_ls_gentb(prjroot, tests):
    """List Generic Testbenches Only."""
    run2console(prjroot, "ls", "-g")
    assert_refdata(test_ls_gentb, prjroot)


def test_ls_pat(prjroot, tests):
    """List Command with Pattern."""
    run2console(prjroot, "ls", "*SomeMod")
    assert_refdata(test_ls_pat, prjroot)


def test_ls_tags(prjroot, tests):
    """List Command with Tags."""
    run2console(prjroot, "ls", "--tag", "intf", "--tag", "ip*")
    assert_refdata(test_ls_tags, prjroot)


def test_ls_tb_dut(prjroot, tests):
    """List Testbenches DUTs."""
    run2console(prjroot, "ls", "tests.*", "-n")
    assert_refdata(test_ls_tb_dut, prjroot)


def test_ls_tb_dut_sub(prjroot, example_simple):
    """List Testbenches DUT with Subs."""
    run2console(prjroot, "ls", "glbl_lib.regf_tb#uart_lib.uart-*", "-n")
    assert_refdata(test_ls_tb_dut_sub, prjroot)


def test_ls_tb_dut_sub_all(prjroot, tests):
    """List Testbenches DUT with Subs, glob."""
    run2console(prjroot, "ls", "*#*-*", "-n")
    assert_refdata(test_ls_tb_dut_sub_all, prjroot)


def test_ls_none(prjroot, tests):
    """List Testbenches DUT with Subs, glob."""
    run2console(prjroot, "ls", "#", exit_code=1)
    assert_refdata(test_ls_none, prjroot)


def test_autocomplete_top(example_simple):
    """Autocompletion for Top."""
    assert len(u.cliutil.auto_top(None, None, "")) > 20
    assert u.cliutil.auto_top(None, None, "gl") == [
        "glbl_lib.clk_gate",
        "glbl_lib.regf",
        "glbl_lib.regf_tb",
    ]
    assert u.cliutil.auto_top(None, None, "glbl_lib.r") == [
        "glbl_lib.regf",
        "glbl_lib.regf_tb",
    ]


def test_autocomplete_path(prjroot):
    """Autocompletion for Path."""
    with chdir(prjroot):
        Path("aaa.txt").touch()
        Path("aab.txt").touch()
        Path("ac.txt").touch()

        assert u.cliutil.auto_path(None, None, "a") == ["aaa.txt", "aab.txt", "ac.txt"]
        assert u.cliutil.auto_path(None, None, "aa") == ["aaa.txt", "aab.txt"]
        assert u.cliutil.auto_path(None, None, "b") == []


def test_toppath(prjroot, example_simple):
    """Check Command."""
    run2console(prjroot, "check", str(example_simple / "uart_lib" / "uart.py"))
    assert_refdata(test_toppath, prjroot)


def test_modinfo(prjroot, example_simple):
    """Modinfo Command."""
    run2console(prjroot, "modinfo", "uart_lib.uart")
    assert_refdata(test_modinfo, prjroot)


def test_modinfos(prjroot, example_simple):
    """Modinfo Command."""
    run2console(prjroot, "modinfo", "*_lib.*", "-S")
    assert_refdata(test_modinfos, prjroot)


def test_modinfos_param(example_param, prjroot):
    """Modinfo Command."""
    run2console(prjroot, "modinfo", "*_lib.*")
    assert_refdata(test_modinfos_param, prjroot)


def test_create(tmp_path):
    """Test Command For The Create Function."""
    with chdir(tmp_path):
        run("create", "-T", "--module", "my_name", "--library", "my_library", "--flavour", "AMod")
    assert_refdata(test_create, tmp_path, replacements=((Path("my_library"), "my_library"),))


def test_create_numbers(tmp_path):
    """Test Create Command With Specified With Numbers."""
    with chdir(tmp_path):
        run("create", "-T", "-m", "my_name2", "-l", "my_library_2", "-F", "AMod")
    assert_refdata(test_create_numbers, tmp_path, replacements=((Path("my_library_2"), "my_library"),))


def test_create_invalid_name(tmp_path):
    """Test Command For The Create Function But To Test Where The Maximum Is."""
    with chdir(tmp_path):
        run(
            "create",
            "-T",
            "--module",
            "my_name_2_previus",
            "--library",
            "my_library_2_previus.py",
            "--flavour",
            "AMod",
            exit_code=1,
        )
    # Check that no file is generated
    assert tuple(tmp_path.glob("*")) == ()


def test_create_regf(tmp_path):
    """Test Create Command With Specified With Numbers."""
    with chdir(tmp_path):
        run("create", "-T", "--module", "my_name_regf", "--library", "my_library", "--regf", "--flavour", "AMod")
    assert_refdata(test_create_regf, tmp_path, replacements=((Path("my_library"), "my_library"),))


def test_create_no_regf(tmp_path):
    """Test Create Command With Specified With Numbers."""
    with chdir(tmp_path):
        run("create", "-T", "--module", "my_name_no_regf", "--library", "my_library", "--no-regf", "--flavour", "AMod")
    assert_refdata(test_create_no_regf, tmp_path, replacements=((Path("my_library"), "my_library"),))


def test_create_descr(tmp_path):
    """Test Create Command With Specified Description."""
    with chdir(tmp_path):
        run(
            "create",
            "-T",
            "--module",
            "my_name_descr",
            "--library",
            "my_library",
            "--descr",
            "My Fancy Module",
            "--flavour",
            "AMod",
        )
    assert_refdata(test_create_descr, tmp_path, replacements=((Path("my_library"), "my_library"),))


@mark.parametrize(
    "flavour",
    [
        "AConfigurableMod",
        "AConfigurableTbMod",
        "AGenericTbMod",
        "AMod",
        "ATailoredMod",
        "ATbMod",
    ],
)
def test_create_flavour(tmp_path, flavour):
    """Test Create Command Flavour."""
    runner = CliRunner()
    with chdir(tmp_path):
        runner.invoke(
            u.cli.ucdp,
            ["create", "-T", "--name", "my_name_flavour", "--library", "my_library", "--flavour", "{flavour}"],
            flavour="\n".join((*flavour, "")),
        )
    assert_refdata(
        test_create_flavour, tmp_path, flavor="-".join(flavour), replacements=((Path("my_library"), "my_library"),)
    )


@mark.parametrize(
    "input",
    [
        ("mod0", "lib_type", "d", "n"),
        ("mod1", "lib_type", "d", "y", "c"),
        ("mod2", "lib_type", "d", "y", "t"),
        ("mod3", "lib_type", "t", "y", "g"),
        ("mod4", "lib_type", "t", "y", "c"),
        ("mod5", "lib_type", "t", "n"),
        ("mod6_tb", "lib_type", "d", "n"),
        ("mod7_tb", "lib_type", "t", "n"),
    ],
)
def test_create_type_questions(tmp_path, input):
    """Test Type Questionnaire."""
    runner = CliRunner()
    with chdir(tmp_path):
        result = runner.invoke(u.cli.ucdp, ["create", "-T"], input="\n".join((*input, "")))
    assert not result.exception
    output_path = tmp_path / "output.txt"
    output_path.write_text(result.output)
    assert_refdata(
        test_create_type_questions, tmp_path, flavor="-".join(input), replacements=((Path("lib_type"), "lib_type"),)
    )


@mark.parametrize("input", [("mod0", "y"), ("mod1", "n")])
def test_create_tb_questions(tmp_path, input):
    """Test tb Questionnaire."""
    runner = CliRunner()
    with chdir(tmp_path):
        result = runner.invoke(
            u.cli.ucdp, ["create", "--library", "lib_tb", "--flavour", "amod"], input="\n".join((*input, ""))
        )
    assert not result.exception
    assert_refdata(
        test_create_tb_questions, tmp_path, flavor="-".join(input), replacements=((Path("lib_tb"), "lib_tb"),)
    )



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Enumeration Types.

An enumeration is a normal base type with a specific mapping of values to a another meaning.

* :any:`EnumItem` - Enumeration item
* :any:`AEnumType` - Standard Enumeration
* :any:`AGlobalEnumType` - A public enumeration which fills up through all instances.
* :any:`DynamicEnumType` - A public enumeration which fills up per instance.
* :any:`EnaType` - Native single bit with `ena` and `dis` enumeration, active-high
* :any:`DisType` - Native single bit with `ena` and `dis` enumeration, low-high
"""

from abc import abstractmethod
from collections.abc import Callable
from typing import Any, ClassVar

from humanfriendly.text import concatenate
from pydantic import model_validator

from .casting import Casting
from .consts import AUTO
from .dict import Dict
from .doc import Doc
from .exceptions import LockError
from .object import Field, Object, PosArgs, PrivateField
from .typebase import AScalarType, BaseScalarType, BaseType


class EnumItem(Object):
    """
    Enumeration NamedObject.

    Args:
        key (int): key value to be mapped.
        value (Any): Mapped value.

    Keyword Args:
        doc (Doc): Documentation Container

    Enumeration items are typically created by :any:`EnumType._add`.
    """

    key: int
    value: Any
    doc: Doc = Doc()

    _posargs: ClassVar[PosArgs] = ("key", "value")

    def __init__(self, key, value, **kwargs):
        super().__init__(key=key, value=value, **kwargs)


EnumItemFilter = Callable[[EnumItem], bool]


class BaseEnumType(BaseScalarType, Dict):
    """Base Type for all Enums."""

    keytype: AScalarType
    valuetype: Any = None
    default: Any = None
    filter_: EnumItemFilter | None = Field(default=None, repr=False)
    _is_locked: bool = PrivateField(default=False)

    def _add(self, key, value, title: str | None = None, descr: str | None = None, comment: str | None = None) -> None:
        """
        Add NamedObject To Enumeration.

        Args:
            key (int): key value to be mapped.
            value: Mapped value.

        Keyword Args:
            title (str): Full Spoken Name.
            descr (str): Documentation Description.
            comment (str): Source Code Comment.

        :meta public:
        """
        if self._is_locked:
            raise LockError(f"{self}: Cannot add item {key!r}={value!r}.")
        items = self._items
        if key is AUTO:
            keys = items.keys()
            key = max(keys) + 1 if keys else 0
        self.keytype.check(key)
        valuetype = self.valuetype
        if valuetype:
            valuetype.check(value)
        if key in items.keys():
            raise ValueError(f"key {key!r} already exists in {self}")
        doc = Doc(title=title, descr=descr, comment=comment)
        enumitem = EnumItem(key, value, doc=doc)
        if not self.filter_ or self.filter_(enumitem):
            items[key] = enumitem

    @property
    def width(self):
        """Width in Bits."""
        return self.keytype.width

    def check(self, value, what="Value"):
        """Check `value`."""
        return self.keytype.check(value, what)

    def encode(self, value, usedefault=False):
        """Encode Value."""
        if usedefault:
            try:
                return self.get_byvalue(value).key
            except ValueError:
                return self.default
        return self.get_byvalue(value).key

    def decode(self, value, usedefault=False):
        """Decode Value."""
        if usedefault:
            try:
                return self.get_bykey(value).value
            except ValueError:
                return self.get_bykey(self.default).value
        else:
            return self.get_bykey(value).value

    @property
    def is_full(self) -> bool:
        """Return `True` if Enumeration Is Fully Encoded."""
        return len(self) == (2 ** int(self.width))

    def get_bykey(self, key) -> EnumItem:
        """Return :any:`EnumItem` with key `key`."""
        item = self.get(key)
        if item is not None:
            return item
        keys = concatenate([repr(item) for item in self.keys()])
        raise ValueError(f"{self} does not contain key {key!r}. Known keys are {keys}.")

    def get_byvalue(self, value) -> EnumItem:
        """Return :any:`EnumItem` with value `value`."""
        for item in self.values():
            if item.value == value:
                return item
        values = concatenate([repr(item.value) for item in self.values()])
        raise ValueError(f"{self} does not contain value {value!r}. Known values are {values}.")

    def get_value(self, key):
        """Return `value` for `key`."""
        return self.get_bykey(key).value

    def get_key(self, value):
        """Return `key` for `value`."""
        return self.get_byvalue(value).key

    def get_hex(self, value=None):
        """Get Hex Value."""
        if value is None:
            value = self.default
        return self.keytype.get_hex(value=value)

    def is_connectable(self, other):
        """Check For Valid Connection To `other`."""
        return (
            isinstance(other, BaseEnumType)
            and self.keytype.is_connectable(other.keytype)
            and self.valuetype == other.valuetype
            and self.keys() == other.keys()
            and all(
                selfitem.value == otheritem.value
                for selfitem, otheritem in zip(self.values(), other.values(), strict=False)
            )
        )

    def __getitem__(self, slice_):
        """Return Slice."""
        return self.keytype[slice_]

    @property
    def min_(self):
        """Minimal Value."""
        return self.keytype.min_

    @property
    def max_(self):
        """Maximal Value."""
        return self.keytype.max_

    @property
    def bits(self):
        """Size in Bits."""
        return self.keytype.bits

    def cast(self, other: BaseType) -> Casting:
        """
        How to cast an input of type `self` from a value of type `other`.

        `self = cast(other)`
        """
        if isinstance(other, BaseEnumType) and self.keytype.is_connectable(other.keytype):
            return [("", "")]

        if isinstance(other, BaseScalarType) and (self.keytype.is_connectable(other) or self.width == other.width):
            return [("", "")]

        return None

    @model_validator(mode="after")
    def __post_init(self) -> "BaseEnumType":
        if self.default is None:
            self.__dict__["default"] = self.keytype.default
        return self

    @abstractmethod
    def _build(self) -> None:
        """Build Type."""



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Module File Information."""

import ucdp as u


class TailMod(u.ATailoredMod):
    """Example Tailored Module."""

    def _build(self):
        self.add_port(u.ClkRstAnType(), "main_i")
        self.add_port(u.UintType(8), "data_i")
        self.add_port(u.UintType(8), "data_o")


class TopMod(u.AMod):
    """Top Module."""

    def _build(self):
        TailMod(self, "u_sub0")
        TailMod(self, "u_sub1")


def test_noparent():
    """Without Parent."""
    mod = TailMod()
    assert mod.modname == "tail"
    assert mod.topmodname == "tail"
    assert mod.libname == "tests"


def test_parent():
    """With Parent."""
    top = TopMod()
    sub0 = top.get_inst("u_sub0")
    assert sub0.modname == "top_sub0"
    assert sub0.topmodname == "top"
    assert sub0.libname == "tests"
    sub1 = top.get_inst("u_sub1")
    assert sub1.modname == "top_sub1"
    assert sub1.topmodname == "top"
    assert sub1.libname == "tests"



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Configuration."""

import re

from hypothesis import assume, given
from hypothesis import strategies as st
from pytest import raises

import ucdp


class MyObject(ucdp.NamedObject):
    """Example Object."""

    arg1: int
    arg2: int


def test_default():
    """Basic Testing."""
    # Create empty Namespace
    namespace = ucdp.Namespace()
    # test basic properties: len(), tuple(), lock(), repr()
    assert len(namespace) == 0
    assert repr(namespace) == "Namespace([])"
    assert tuple(namespace) == ()
    assert namespace.is_locked is False
    namespace.lock()
    assert namespace.is_locked is True


@given(
    int1=st.integers(),
    int2=st.integers(),
    int3=st.integers(),
    int4=st.integers(),
    int5=st.integers(),
    int6=st.integers(),
    str1=st.text(),
    str2=st.text(),
    str3=st.text(),
)
def test_init(int1, int2, int3, int4, int5, int6, str1, str2, str3):
    """Namespace with Initialization Items."""
    assume(str1 != str2)
    assume(str2 != str3)
    assume(str1 != str3)

    myobject1 = MyObject(name=str1, arg1=int1, arg2=int2)
    myobject2 = MyObject(name=str2, arg1=int3, arg2=int4)
    myobject3 = MyObject(name=str3, arg1=int5, arg2=int6)
    assert myobject1.name == str1
    assert myobject1.arg1 == int1
    assert myobject1.arg2 == int2
    assert myobject2.name == str2
    assert myobject2.arg1 == int3
    assert myobject2.arg2 == int4
    assert myobject3.name == str3
    assert myobject3.arg1 == int5
    assert myobject3.arg2 == int6

    # Create Namespace with Elements in Constructor
    namespace = ucdp.Namespace([myobject1, myobject2, myobject3])
    # test basic properties: len(), tuple(), lock(), repr()
    assert namespace[str1] == myobject1
    assert namespace[str2] == myobject2
    assert namespace[str3] == myobject3
    assert len(namespace) == 3
    assert repr(namespace) == f"Namespace([{myobject1!r}, {myobject2!r}, {myobject3!r}])"
    assert tuple(namespace) == (myobject1, myobject2, myobject3)
    assert namespace.is_locked is False
    namespace.lock()
    assert namespace.is_locked is True


@given(
    int1=st.integers(),
    int2=st.integers(),
    int3=st.integers(),
    int4=st.integers(),
    int5=st.integers(),
    int6=st.integers(),
    str1=st.text(),
    str2=st.text(),
)
def test_add(int1, int2, int3, int4, int5, int6, str1, str2):
    """Namespace add() Method."""
    assume(str1 != str2)

    myobject1 = MyObject(name=str1, arg1=int1, arg2=int2)
    myobject2 = MyObject(name=str2, arg1=int3, arg2=int4)
    myobject3 = MyObject(name=str1, arg1=int5, arg2=int6)

    # Create empty Namespace
    namespace = ucdp.Namespace([])
    assert len(namespace) == 0

    # Add new Object
    namespace.add(myobject1)
    assert namespace[str1] == myobject1
    assert len(namespace) == 1
    assert repr(namespace) == f"Namespace([{myobject1!r}])"
    assert tuple(namespace) == (myobject1,)

    # Add 2nd Object
    namespace.add(myobject2)
    assert namespace[str1] == myobject1
    assert namespace[str2] == myobject2
    assert len(namespace) == 2
    assert repr(namespace) == f"Namespace([{myobject1!r}, {myobject2!r}])"
    assert tuple(namespace) == (myobject1, myobject2)

    # Try to add Object with same name
    with raises(ucdp.exceptions.DuplicateError) as excep:
        namespace.add(myobject3)
    assert str(excep.value) == f"Name '{str1}' already taken by {namespace[str1]!r}"
    assert namespace[str1] == myobject1
    assert namespace[str2] == myobject2
    assert len(namespace) == 2
    assert repr(namespace) == f"Namespace([{myobject1!r}, {myobject2!r}])"
    assert tuple(namespace) == (myobject1, myobject2)


@given(
    int1=st.integers(),
    int2=st.integers(),
    int3=st.integers(),
    int4=st.integers(),
    str1=st.text(),
    str2=st.text(),
    str3=st.text(),
)
def test_get(int1, int2, int3, int4, str1, str2, str3):
    """Namespace get() Method."""
    assume(str1 != str2)
    assume(str2 != str3)
    assume(str1 != str3)

    myobject1 = MyObject(name=str1, arg1=int1, arg2=int2)
    myobject2 = MyObject(name=str2, arg1=int3, arg2=int4)

    # Create Namespace with Elements in Constructor
    namespace = ucdp.Namespace([myobject1, myobject2])
    assert namespace[myobject1.name] == myobject1
    assert namespace[myobject2.name] == myobject2
    with raises(KeyError) as excep:
        namespace[str3]
    assert str(excep.value) == f"{str3!r}"


@given(
    int1=st.integers(),
    int2=st.integers(),
    int3=st.integers(),
    int4=st.integers(),
    str1=st.text(),
    str2=st.text(),
    str3=st.text(),
)
def test_get_dym(int1, int2, int3, int4, str1, str2, str3):
    """Namespace get_dym() Method."""
    assume(str1 != str2)
    assume(str2 != str3)
    assume(str1 != str3)

    myobject1 = MyObject(name=str1, arg1=int1, arg2=int2)
    myobject2 = MyObject(name=str2, arg1=int3, arg2=int4)

    # Create Namespace with Elements in Constructor
    namespace = ucdp.Namespace([myobject1, myobject2])
    assert namespace.get_dym(myobject1.name) == myobject1
    assert namespace.get_dym(myobject2.name) == myobject2
    with raises(ValueError) as excep:
        namespace.get_dym(str3)

    tmp_dict = {myobject1.name: myobject1, myobject2.name: myobject2}
    dym = ucdp.nameutil.didyoumean(str3, tmp_dict.keys(), known=True)
    assert str(excep.value) == f"{str3!r}.{dym}"


@given(
    int1=st.integers(),
    int2=st.integers(),
    int3=st.integers(),
    int4=st.integers(),
    str1=st.text(),
    str2=st.text(),
    str3=st.text(),
)
def test_getitem(int1, int2, int3, int4, str1, str2, str3):
    """Namespace set item Method."""
    assume(str1 != str2)
    assume(str2 != str3)
    assume(str1 != str3)

    myobject1 = MyObject(name=str1, arg1=int1, arg2=int2)
    myobject2 = MyObject(name=str2, arg1=int3, arg2=int4)

    # Create Namespace with Elements in Constructor
    namespace = ucdp.Namespace([myobject1, myobject2])
    assert namespace[myobject1.name] == myobject1
    assert namespace[myobject2.name] == myobject2
    tmp_dict = {myobject1.name: myobject1, myobject2.name: myobject2}
    assert namespace.items() == tmp_dict.items()
    assert namespace.keys() == tmp_dict.keys()
    assert repr(namespace.values()) == repr(tmp_dict.values())


@given(
    int1=st.integers(),
    int2=st.integers(),
    int3=st.integers(),
    int4=st.integers(),
    int5=st.integers(),
    int6=st.integers(),
    str1=st.text(),
    str2=st.text(),
    str3=st.text(),
)
def test_setitem(int1, int2, int3, int4, int5, int6, str1, str2, str3):
    """Namespace set item Method."""
    assume(str1 != str2)
    assume(str1 != str3)
    assume(str2 != str3)

    myobject1 = MyObject(name=str1, arg1=int1, arg2=int2)
    myobject2 = MyObject(name=str2, arg1=int3, arg2=int4)
    myobject3 = MyObject(name=str1, arg1=int5, arg2=int6)

    # Create empty Namespace
    namespace = ucdp.Namespace([])
    assert len(namespace) == 0

    # Add new Object
    namespace[myobject1.name] = myobject1
    assert namespace[str1] == myobject1
    assert len(namespace) == 1
    assert repr(namespace) == f"Namespace([{myobject1!r}])"
    assert tuple(namespace) == (myobject1,)

    # Try to add same Object
    with raises(ucdp.exceptions.DuplicateError) as excep:
        namespace[myobject1.name] = myobject1
    assert str(excep.value) == f"{myobject1!r} already exists"

    # Try to add 2nd Object with not matching name
    with raises(ValueError) as excep:
        namespace[str3] = myobject2
    assert str(excep.value) == f"{myobject2} with must be stored at name '{myobject2.name}' not at '{str3}'"

    # Add 2nd Object
    namespace[myobject2.name] = myobject2
    assert namespace[str1] == myobject1
    assert namespace[str2] == myobject2
    assert len(namespace) == 2
    assert repr(namespace) == f"Namespace([{myobject1!r}, {myobject2!r}])"
    assert tuple(namespace) == (myobject1, myobject2)

    # Try to add Object with same name
    with raises(ucdp.exceptions.DuplicateError) as excep:
        namespace[myobject3.name] = myobject3
    assert str(excep.value) == f"Name '{str1}' already taken by {namespace[str1]!r}"
    assert namespace[str1] == myobject1
    assert namespace[str2] == myobject2
    assert len(namespace) == 2
    assert repr(namespace) == f"Namespace([{myobject1!r}, {myobject2!r}])"
    assert tuple(namespace) == (myobject1, myobject2)

    # Try to add Object with same name with update()
    with raises(ucdp.exceptions.DuplicateError) as excep:
        namespace.update({myobject3.name: myobject3})
    assert str(excep.value) == f"Name '{str1}' already taken by {namespace[str1]!r}"
    assert namespace[str1] == myobject1
    assert namespace[str2] == myobject2
    assert len(namespace) == 2
    assert repr(namespace) == f"Namespace([{myobject1!r}, {myobject2!r}])"
    assert tuple(namespace) == (myobject1, myobject2)

    # Add Object with update()
    myobject3 = MyObject(name=str3, arg1=int5, arg2=int6)
    namespace.update({myobject3.name: myobject3})
    assert namespace[str1] == myobject1
    assert namespace[str2] == myobject2
    assert namespace[str3] == myobject3
    assert len(namespace) == 3
    assert repr(namespace) == f"Namespace([{myobject1!r}, {myobject2!r}, {myobject3!r}])"
    assert tuple(namespace) == (myobject1, myobject2, myobject3)


@given(int1=st.integers(), int2=st.integers(), str1=st.text())
def test_delitem(int1, int2, str1):
    """Namespace set item Method."""
    # Create Namespace with Elements in Constructor
    namespace = ucdp.Namespace([MyObject(name=str1, arg1=int1, arg2=int2)])

    assert len(namespace) == 1

    with raises(TypeError) as excep:
        del namespace[str1]
    assert str(excep.value) == f"It is forbidden to remove {str1!r}."

    with raises(TypeError) as excep:
        namespace.pop(str1)
    assert str(excep.value) == "It is forbidden to remove any item."

    with raises(TypeError) as excep:
        namespace.popitem()
    assert str(excep.value) == "It is forbidden to remove any item."

    assert len(namespace) == 1


@given(
    int1=st.integers(),
    int2=st.integers(),
    int3=st.integers(),
    int4=st.integers(),
    str1=st.text(),
    str2=st.text(),
)
def test_lock(int1, int2, int3, int4, str1, str2):
    """Namespace lock Method."""
    assume(str1 != str2)

    myobject1 = MyObject(name=str1, arg1=int1, arg2=int2)
    myobject2 = MyObject(name=str2, arg1=int3, arg2=int4)

    # Create empty Namespace
    namespace = ucdp.Namespace([])
    assert len(namespace) == 0

    # Add new Object
    namespace.add(myobject1)
    assert len(namespace) == 1
    assert namespace.is_locked is False
    namespace.lock()
    assert namespace.is_locked is True

    msg = "Namespace is already locked. Cannot lock again."
    with raises(ucdp.LockError, match=re.escape(msg)):
        namespace.lock()

    msg = "Namespace is already locked. Cannot add items anymore."
    with raises(ucdp.LockError, match=re.escape(msg)):
        namespace.add(myobject2)

    msg = "Namespace is already locked. Cannot add items anymore."
    with raises(ucdp.LockError, match=re.escape(msg)):
        namespace.update({myobject2.name: myobject2})

    assert len(namespace) == 1


def test_ior():
    """Namespace IOR-Operator."""
    myobject1 = ucdp.NamedObject(name="str1")
    myobject2 = ucdp.NamedObject(name="str2")
    myobject3 = ucdp.NamedObject(name="str3")
    namespace = ucdp.Namespace()
    namespace.add(myobject1)
    assert tuple(namespace) == (myobject1,)

    namespace |= {"str2": myobject2}
    assert tuple(namespace) == (myobject1, myobject2)

    namespace.lock()
    assert tuple(namespace) == (myobject1, myobject2)

    msg = "Namespace is already locked. Cannot add items anymore."
    with raises(ucdp.LockError, match=re.escape(msg)):
        namespace |= {"str3": myobject3}

    assert tuple(namespace) == (myobject1, myobject2)


def test_or():
    """Namespace IOR-Operator."""
    myobject1 = ucdp.NamedObject(name="str1")
    myobject2 = ucdp.NamedObject(name="str2")
    myobject3 = ucdp.NamedObject(name="str3")
    namespace = ucdp.Namespace()
    namespace.add(myobject1)
    assert tuple(namespace) == (myobject1,)

    namespace2 = namespace | {"str2": myobject2}
    assert tuple(namespace) == (myobject1,)
    assert tuple(namespace2) == (myobject1, myobject2)

    namespace2.lock()
    assert tuple(namespace2) == (myobject1, myobject2)

    msg = "Namespace is already locked. Cannot add items anymore."
    with raises(ucdp.LockError, match=re.escape(msg)):
        namespace2 | {"str3": myobject3}

    assert tuple(namespace2) == (myobject1, myobject2)


def test_set_default():
    """Namespace IOR-Operator."""
    myobject1 = ucdp.NamedObject(name="str1")
    myobject2 = ucdp.NamedObject(name="str2")
    myobject3 = ucdp.NamedObject(name="str3")
    namespace = ucdp.Namespace()
    assert namespace.set_default("str1", myobject1) is myobject1
    assert tuple(namespace) == (myobject1,)
    assert namespace.set_default("str1", myobject1) is myobject1
    assert tuple(namespace) == (myobject1,)

    msg = "NamedObject(name='str3') with must be stored at name 'str3' not at 'str2'"
    with raises(ValueError, match=re.escape(msg)):
        namespace.set_default("str2", myobject3)

    assert tuple(namespace) == (myobject1,)

    namespace.lock()

    msg = "Namespace is already locked. Cannot add items anymore."
    with raises(ValueError, match=re.escape(msg)):
        namespace.set_default("str2", myobject2)

    assert tuple(namespace) == (myobject1,)



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Multiplexer."""

import re

from pytest import raises

from ucdp import DOWN, UP, Const, Slice, UintType


def test_str():
    """Slice from String."""
    assert Slice(left="5").bits == "5"
    assert Slice(left="[5:7]").bits == "5:7"
    assert Slice(left="[7:5]").bits == "7:5"
    with raises(ValueError, match=re.escape("'right' must be None")):
        Slice(left="6", right=4)
    with raises(ValueError, match=re.escape("'width' must be None")):
        Slice(left="6", width=4)
    with raises(ValueError, match=re.escape("Invalid Slice Specification ''")):
        Slice(left="")


def test_width():
    """Slice with Width."""
    assert Slice(width=2).bits == "1:0"
    assert Slice(left=4, width=2).bits == "4:3"
    assert Slice(right=4, width=2).bits == "5:4"
    with raises(ValueError, match=re.escape("'left', 'right' AND 'width' given, this is one too much")):
        Slice(left=1, right=2, width=3)


def test_up():
    """Slice Upwards."""
    assert Slice(width=2, direction=UP).bits == "0:1"
    assert Slice(width=2, left=2, direction=UP).bits == "2:3"
    assert Slice(width=2, right=2, direction=UP).bits == "1:2"


def test_down():
    """Slice Downwards."""
    assert Slice(width=2, direction=DOWN).bits == "1:0"
    assert Slice(width=2, left=2, direction=DOWN).bits == "2:1"
    assert Slice(width=2, right=2, direction=DOWN).bits == "3:2"


def test_const():
    """Const."""
    const4 = Const(UintType(8, default=4), "const4")

    assert (
        Slice(width=const4).bits
        == "Op(Const(UintType(8, default=4), 'const4'), '-', ConstExpr(UintType(8, default=1))):0"
    )
    assert (
        Slice(width=const4, right=0).bits
        == "Op(Const(UintType(8, default=4), 'const4'), '-', ConstExpr(UintType(8, default=1))):0"
    )
    assert (
        Slice(width=const4, right=2).bits
        == "Op(Op(Const(UintType(8, default=4), 'const4'), '-', ConstExpr(UintType(8, default=1))), '+', "
        "ConstExpr(UintType(8, default=2))):2"
    )

    assert Slice(right=const4).bits == "const4"
    assert Slice(right=const4, width=1).bits == "const4"
    assert (
        Slice(right=const4, width=3).bits
        == "Op(ConstExpr(UintType(8, default=2)), '+', Const(UintType(8, default=4), 'const4')):const4"
    )

    assert Slice(left=const4).bits == "const4"
    assert Slice(left=const4, width=1).bits == "const4"
    assert (
        Slice(left=const4, width=3).bits
        == "const4:Op(Const(UintType(8, default=4), 'const4'), '-', ConstExpr(UintType(8, default=2)))"
    )



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Base Hardware Module.

[BaseMod][ucdp.modbase.BaseMod] defines the base interface which is **common to all hardware modules**.
"""

import warnings
from abc import abstractmethod
from functools import cached_property
from inspect import getmro
from typing import Annotated, Any, ClassVar, Literal, Optional, TypeAlias, Union, no_type_check

from aligntext import align
from caseconverter import snakecase
from pydantic import PlainValidator
from uniquer import uniquetuple

from .assigns import Assigns, Drivers, Note, Source
from .baseclassinfo import get_baseclassinfos
from .clkrel import ClkRel
from .clkrelbase import BaseClkRel
from .const import Const
from .consts import UPWARDS
from .define import Define, Defines
from .doc import Doc
from .docutil import doc_from_type
from .exceptions import LockError
from .expr import BoolOp, Expr
from .exprparser import ExprParser, Parseable, cast_booltype
from .flipflop import FlipFlop
from .ident import Ident, Idents
from .ifdef import Ifdefs, cast_ifdefs, resolve_ifdefs
from .iterutil import namefilter
from .logging import LOGGER
from .modref import ModRef, get_modclsname
from .modutil import get_modbaseinfos
from .mux import Mux
from .namespace import Namespace
from .nameutil import join_names, split_prefix
from .object import Field, NamedObject, Object, PrivateField, computed_field
from .orientation import FWD, IN, Direction, Orientation
from .param import Param
from .routepath import Routeables, RoutePath, parse_routepaths
from .signal import BaseSignal, Port, Signal
from .typebase import BaseType
from .typedescriptivestruct import DescriptiveStructType
from .typestruct import StructItem

ModTags: TypeAlias = set[str]
RoutingError: TypeAlias = Literal["error", "warn", "ignore"]


def _to_defines(value):
    if value is None:
        return Namespace()
    if isinstance(value, Namespace):
        value.lock()
        return value
    if isinstance(value, dict):
        defines = Namespace()
        for key, val in value.items():
            defines.add(Define(key, value=val))
        return defines
    raise ValueError(value)


class BaseMod(NamedObject):
    """
    Hardware Module.

    Args:
        parent: Parent Module. `None` by default for top module.
        name: Instance name. Required if parent is provided.

    Keyword Args:
        title (str): Title
        descr (str): Description
        comment (str): Comment
        paramdict (dict): Parameter values for this instance.
    """

    filelists: ClassVar[Any] = ()
    tags: ClassVar[ModTags] = ModTags()

    parent: Optional["BaseMod"] = None
    paramdict: dict = Field(default_factory=dict, repr=False)

    title: str | None = None
    descr: str | None = None
    comment: str | None = None

    has_hiername: bool = True

    virtual: bool = False

    # private

    drivers: Drivers = Field(default_factory=Drivers, init=False, repr=False)
    defines: Annotated[Defines, PlainValidator(_to_defines)] = Field(default_factory=Defines)
    namespace: Idents = Field(default_factory=Idents, init=False, repr=False)
    params: Idents = Field(default_factory=Idents, init=False, repr=False)
    ports: Idents = Field(default_factory=Idents, init=False, repr=False)
    portssignals: Idents = Field(default_factory=Idents, init=False, repr=False)
    insts: Namespace = Field(default_factory=Namespace, init=False, repr=False)

    _has_build_dep: bool = PrivateField(default=False)
    _has_build_final: bool = PrivateField(default=False)

    __is_locked: bool = PrivateField(default=False)
    __instcons: dict[str, tuple[Assigns, ExprParser]] = PrivateField(default_factory=dict)
    __flipflops: dict[int, FlipFlop] = PrivateField(default_factory=dict)
    __muxes: Namespace = PrivateField(default_factory=Namespace)
    __parents = PrivateField(default_factory=list)

    def __init__(self, parent: Optional["BaseMod"] = None, name: str | None = None, defines=None, **kwargs):
        cls = self.__class__
        if not cls.__name__.endswith("Mod"):
            raise NameError(f"Name of {cls} MUST end with 'Mod'")
        if not name:
            if parent:
                raise ValueError("'name' is required for sub modules.")
            name = snakecase(cls.__name__.removesuffix("Mod"))
        super().__init__(parent=parent, name=name, defines=defines, **kwargs)  # type: ignore[call-arg]

    @property
    def doc(self) -> Doc:
        """Documentation."""
        return Doc(title=self.title, descr=self.descr, comment=self.comment)

    @property
    def basename(self) -> str:
        """Base Name Derived From Instance."""
        return split_prefix(self.name)[1]

    @property
    @abstractmethod
    def modname(self) -> str:
        """Module Name."""

    @property
    @abstractmethod
    def topmodname(self) -> str:
        """Top Module Name."""

    @property
    def libname(self) -> str:
        """Library Name."""
        return self.libpath.name

    @property
    @abstractmethod
    def libpath(self) -> str:
        """Library Path."""

    @cached_property
    def qualname(self) -> str:
        """Qualified Name (Library Name + Module Name)."""
        return f"{self.libname}.{self.modname}"

    @cached_property
    def basequalnames(self) -> tuple[str, ...]:
        """Qualified Name (Library Name + Module Name) of Base Modules."""
        return uniquetuple(f"{bci.libname}.{bci.modname}" for bci in get_modbaseinfos(self))

    @classmethod
    def get_modref(cls, minimal: bool = False) -> ModRef:
        """Python Class Reference."""
        bci = next(get_baseclassinfos(cls))
        modclsname = bci.clsname if not minimal or bci.clsname != get_modclsname(bci.modname) else None

        return ModRef(
            libname=bci.libname,
            modname=bci.modname,
            modclsname=modclsname,
        )

    @classmethod
    def get_basemodrefs(cls) -> tuple[ModRef, ...]:
        """Python Class Reference."""
        return tuple(
            ModRef(
                libname=bci.libname,
                modname=bci.modname,
                modclsname=bci.clsname,
            )
            for bci in get_modbaseinfos(cls)
        )

    @property
    def hiername(self) -> str:
        """Hierarchical Name."""
        mod: BaseMod | None = self
        names: list[str] = []
        while mod is not None:
            if mod.has_hiername:
                names.insert(0, split_prefix(mod.name)[1])
            mod = mod.parent
        return join_names(*names)

    @property
    @abstractmethod
    def is_tb(self) -> bool:
        """Determine if module belongs to Testbench or Design."""

    @property
    def path(self) -> tuple["BaseMod", ...]:
        """Path."""
        path = [self]
        parent = self.parent
        while parent:
            path.insert(0, parent)
            parent = parent.parent
        return tuple(path)

    @property
    def inst(self) -> str:
        """Path String."""
        parts: list[str] = []
        mod = self
        while mod.parent:
            parts.insert(0, mod.name)
            mod = mod.parent
        parts.insert(0, mod.modname)
        return "/".join(parts)

    @computed_field
    @cached_property
    def assigns(self) -> Assigns:
        """Assignments."""
        return Assigns(targets=self.portssignals, sources=self.namespace, drivers=self.drivers)

    @computed_field
    @cached_property
    def parser(self) -> ExprParser:
        """Expression Parser."""
        return ExprParser(namespace=self.namespace, context=str(self))

    @computed_field
    @cached_property
    def _router(self) -> "Router":
        """Router."""
        return Router(mod=self)

    @property
    def parents(self) -> tuple["BaseMod", ...]:
        """Parents."""
        return tuple(self.__parents)

    @classmethod
    def build_top(cls, **kwargs) -> "BaseMod":
        """
        Build Top Instance.

        Return module as top module.
        """
        return cls(**kwargs)

    def add_param(
        self,
        arg: BaseType | Param,
        name: str | None = None,
        title: str | None = None,
        descr: str | None = None,
        comment: str | None = None,
        ifdef: str | None = None,
        ifdefs: Ifdefs | str | None = None,
        exist_ok: bool = False,
    ) -> Param | None:
        """
        Add Module Parameter (:any:`Param`).

        Args:
            arg: Type or Parameter
            name: Name. Mandatory if arg is a Type.

        Keyword Args:
            title: Full Spoken Name.
            descr: Documentation Description.
            comment: Source Code Comment.
            ifdef: IFDEF pragma. Obsolete.
            ifdefs: IFDEFs pragmas.
            exist_ok: Do not complain about already existing item
        """
        if ifdef:
            warnings.warn(
                "add_param(..., ifdef=...) is obsolete, please use ifdefs=", category=DeprecationWarning, stacklevel=1
            )
        ifdefs = cast_ifdefs(ifdefs or ifdef)
        rifdefs = resolve_ifdefs(self.defines, ifdefs)
        if rifdefs is None:
            return None
        if isinstance(arg, Param):
            value = self.paramdict.pop(arg.name, None)
            param: Param = arg.new(value=value)
            assert name is None
        else:
            type_: BaseType = arg
            doc = doc_from_type(type_, title=title, descr=descr, comment=comment)
            value = self.paramdict.pop(name, None)
            param = Param(type_=type_, name=name, doc=doc, ifdefs=rifdefs, value=value)
        if self.__is_locked:
            raise LockError(f"{self}: Cannot add parameter {name!r}.")
        self.namespace.add(param, exist_ok=exist_ok)
        self.params.add(param, exist_ok=exist_ok)
        return param

    def add_const(
        self,
        arg: BaseType | Const,
        name: str | None = None,
        title: str | None = None,
        descr: str | None = None,
        comment: str | None = None,
        ifdef: str | None = None,
        ifdefs: Ifdefs | str | None = None,
        exist_ok: bool = False,
    ) -> Const | None:
        """
        Add Module Internal Constant (:any:`Const`).

        Args:
            arg: Type or Parameter
            name: Name. Mandatory if arg is a Type.

        Keyword Args:
            title: Full Spoken Name.
            descr: Documentation Description.
            comment: Source Code Comment.
            ifdef: IFDEF pragma. Obsolete.
            ifdefs: IFDEFs pragmas.
            exist_ok: Do not complain about already existing item
        """
        if ifdef:
            warnings.warn(
                "add_const(..., ifdef=...) is obsolete, please use ifdefs=", category=DeprecationWarning, stacklevel=1
            )
        ifdefs = cast_ifdefs(ifdefs or ifdef)
        rifdefs = resolve_ifdefs(self.defines, ifdefs)
        if rifdefs is None:
            return None
        if isinstance(arg, Const):
            const: Const = arg
            assert name is None
        else:
            type_: BaseType = arg
            doc = doc_from_type(type_, title=title, descr=descr, comment=comment)
            const = Const(type_=type_, name=name, doc=doc, ifdefs=rifdefs)
        if self.__is_locked:
            raise LockError(f"{self}: Cannot add constant {name!r}.")
        self.namespace.add(const, exist_ok=exist_ok)
        return const

    def add_type_consts(self, type_: BaseType, exist_ok: bool = False, only=None, name=None, item_suffix="e"):
        """
        Add description of `type_` as local parameters.

        Args:
            type_: Type to be described.

        Keyword Args:
            exist_ok (bool): Do not complain, if parameter already exists.
            name (str): Name of the local parameter. Base on `type_` name by default.
            only (str): Limit parameters to these listed in here, separated by ';'
            item_suffix (str): Enumeration Suffix.
        """
        name = name or snakecase(type_.__class__.__name__.removesuffix("Type"))
        if only:
            patfilter = namefilter(only)

            def filter_(item: StructItem):
                return patfilter(item.name)

            type_ = DescriptiveStructType(type_, filter_=filter_, enumitem_suffix=item_suffix)
        else:
            type_ = DescriptiveStructType(type_, enumitem_suffix=item_suffix)
        self.add_const(type_, name, exist_ok=exist_ok, title=type_.title, descr=type_.descr, comment=type_.comment)

    def add_port(
        self,
        type_: BaseType,
        name: str,
        direction: Direction | None = None,
        title: str | None = None,
        descr: str | None = None,
        comment: str | None = None,
        ifdef: str | None = None,
        ifdefs: Ifdefs | str | None = None,
        route: Routeables | None = None,
        clkrel: str | Port | BaseClkRel | None = None,
    ) -> Port | None:
        """
        Add Module Port (:any:`Port`).

        Args:
            type_: Type.
            name: Name.

        Keyword Args:
            direction: Signal Direction. Automatically detected if `name` ends with '_i', '_o', '_io'.
            title: Full Spoken Name.
            descr: Documentation Description.
            comment: Source Code Comment. Default is 'title'
            ifdef: IFDEF pragma. Obsolete.
            ifdefs: IFDEFs pragmas.
            route: Routes (iterable or string separated by ';')
            clkrel: Clock relation.
        """
        if ifdef:
            warnings.warn(
                "add_port(..., ifdef=...) is obsolete, please use ifdefs=", category=DeprecationWarning, stacklevel=1
            )
        doc = doc_from_type(type_, title, descr, comment)
        if direction is None:
            direction = Direction.from_name(name) or IN
        if clkrel is not None:
            clkrel = self._resolve_clkrel(clkrel)
        ifdefs = cast_ifdefs(ifdefs or ifdef)
        rifdefs = resolve_ifdefs(self.defines, ifdefs)
        if rifdefs is None:
            return None
        port = Port(type_, name, direction=direction, doc=doc, ifdefs=rifdefs, clkrel=clkrel)
        if self.__is_locked:
            raise LockError(f"{self}: Cannot add port {name!r}.")
        self.namespace[name] = port
        self.portssignals[name] = port
        self.ports[name] = port
        for routepath in parse_routepaths(route):
            self._router.add(RoutePath(expr=port), routepath)
        return port

    def _resolve_clkrel(self, clkrel: str | Port | BaseClkRel) -> BaseClkRel:
        if isinstance(clkrel, BaseClkRel):
            return clkrel
        if isinstance(clkrel, BaseSignal):
            return ClkRel(clk=clkrel)
        if isinstance(clkrel, str):
            port = self.ports[clkrel]
            return ClkRel(clk=port)
        raise ValueError(f"Invalid {clkrel=}")

    def add_signal(
        self,
        type_: BaseType,
        name: str,
        direction: Orientation = FWD,
        title: str | None = None,
        descr: str | None = None,
        comment: str | None = None,
        ifdef: str | None = None,
        ifdefs: Ifdefs | str | None = None,
        route: Routeables | None = None,
        clkrel: str | Port | BaseClkRel | None = None,
    ) -> Signal | None:
        """
        Add Module Internal Signal (:any:`Signal`).

        Args:
            type_: Type.
            name: Name.

        Keyword Args:
            direction: Signal Direction. Automatically detected if `name` ends with '_i', '_o', '_io'.
            title: Full Spoken Name.
            descr: Documentation Description.
            comment: Source Code Comment. Default is 'title'
            ifdef: IFDEF pragma. Obsolete.
            ifdefs: IFDEFs pragmas.
            route: Routes (iterable or string separated by ';')
            clkrel: Clock relation.
        """
        if ifdef:
            warnings.warn(
                "add_signal(..., ifdef=...) is obsolete, please use ifdefs=", category=DeprecationWarning, stacklevel=1
            )
        doc = doc_from_type(type_, title, descr, comment)
        if clkrel is not None:
            clkrel = self._resolve_clkrel(clkrel)
        ifdefs = cast_ifdefs(ifdefs or ifdef)
        rifdefs = resolve_ifdefs(self.defines, ifdefs)
        if rifdefs is None:
            return None
        signal = Signal(type_, name, direction=direction, doc=doc, ifdefs=rifdefs, clkrel=clkrel)
        if self.__is_locked:
            raise LockError(f"{self}: Cannot add signal {name!r}.")
        self.namespace[name] = signal
        self.portssignals[name] = signal
        for routepath in parse_routepaths(route):
            self._router.add(RoutePath(expr=signal), routepath)
        return signal

    def add_port_or_signal(
        self,
        type_: BaseType,
        name: str,
        direction: Direction | None = None,
        title: str | None = None,
        descr: str | None = None,
        comment: str | None = None,
        ifdef: str | None = None,
        ifdefs: Ifdefs | str | None = None,
        route: Routeables | None = None,
        clkrel: str | Port | BaseClkRel | None = None,
    ) -> BaseSignal | None:
        """
        Add Module Port (:any:`Port`) or Signal (:any:`Signal`) depending on name.

        Args:
            type_: Type.
            name: Name.

        Keyword Args:
            direction: Signal Direction. Automatically detected if `name` ends with '_i', '_o', '_io'.
            title: Full Spoken Name.
            descr: Documentation Description.
            comment: Source Code Comment. Default is 'title'
            ifdef: IFDEF pragma. Obsolete.
            ifdefs: IFDEFs pragmas.
            route: Routes (iterable or string separated by ';')
            clkrel: Clock relation.
        """
        if ifdef:
            warnings.warn(
                "add_port_or_signal(..., ifdef=...) is obsolete, please use ifdefs=",
                category=DeprecationWarning,
                stacklevel=1,
            )
        ifdefs = cast_ifdefs(ifdefs or ifdef)
        if direction is None:
            direction = Direction.from_name(name)
        if direction is None:
            return self.add_signal(
                type_,
                name,
                title=title,
                descr=descr,
                comment=comment,
                ifdefs=ifdefs,
                route=route,
                clkrel=clkrel,
            )
        return self.add_port(
            type_,
            name,
            direction=direction,
            title=title,
            descr=descr,
            comment=comment,
            ifdefs=ifdefs,
            route=route,
            clkrel=clkrel,
        )

    def set_parent(self, parent: "BaseMod") -> None:
        """
        Set Parent.

        Do not use this method, until you really really really know what you do.
        """
        self.__parents.append(parent)

    def assign(
        self,
        target: Parseable,
        source: Parseable | Note,
        cast: bool = False,
        overwrite: bool = False,
    ):
        """
        Assign `target` to `source`.

        The assignment is done **without** routing.

        Args:
            target: Target to be driven. Must be known within this module.
            source: Source driving target. Must be known within this module.

        Keyword Args:
            cast (bool): Cast. `False` by default.
            overwrite (bool): Overwrite existing assignment.
            filter_ (str, Callable): Target names or function to filter target identifiers.
        """
        if self.__is_locked:
            raise LockError(f"{self}: Cannot add assign '{source}' to '{target}'.")
        parser = self.parser
        assigntarget: BaseSignal = parser.parse(target, only=BaseSignal)  # type: ignore[assignment]
        assignsource: Source = parser.parse_note(source, only=Source)  # type: ignore[arg-type]
        self.assigns.set(assigntarget, assignsource, cast=cast, overwrite=overwrite)

    def add_inst(self, inst: "BaseMod") -> None:
        """
        Add Submodule `inst`.

        Args:
            inst: Instance.
        """
        if self.__is_locked:
            raise LockError(f"{self}: Cannot add instance '{inst}'.")
        inst.set_parent(self)
        self.insts.add(inst)  # type: ignore[arg-type]
        assigns = Assigns(targets=inst.ports, sources=self.namespace, drivers=Drivers(), inst=True)
        parser = ExprParser(namespace=inst.ports, context=str(inst))
        self.__instcons[inst.name] = assigns, parser

    def get_inst(self, inst_or_name: Union["BaseMod", str]) -> "BaseMod":
        """
        Get Module Instance.
        """
        if not isinstance(inst_or_name, str):
            try:
                return self.insts[inst_or_name.name]
            except KeyError:
                raise ValueError(f"{inst_or_name} is not a sub-module of {self}") from None
        inst = self
        for part in inst_or_name.split("/"):
            if part == UPWARDS:
                if inst.parent is None:
                    raise ValueError(f"{self}: {inst} has no parent.")
                inst = inst.parent
            else:
                try:
                    inst = inst.insts.get_dym(part)  # type: ignore[assignment]
                except ValueError as exc:
                    raise ValueError(f"{self} has no sub-module {exc}") from None
        return inst

    def set_instcon(
        self,
        inst: Union["BaseMod", str],
        port: Parseable,
        expr: Parseable,
        cast: bool = False,
        overwrite: bool = False,
    ):
        """
        Connect `port` of `inst` to `expr` without routing.

        The assignment is done **without** routing.

        Args:
            inst: Module Instance
            port: Port to be connected. Must be known within module instance.
            expr: Expression. Must be known within this module.

        Keyword Args:
            cast: Cast. `False` by default.
            overwrite: Overwrite existing assignment.
        """
        if self.__is_locked:
            raise LockError(f"{self}: Cannot connect '{port}' of'{inst}' to '{expr}'.")
        mod: BaseMod = self.get_inst(inst)
        assigns, parser = self.__instcons[mod.name]
        assigntarget: BaseSignal = parser.parse(port, only=BaseSignal)  # type: ignore[assignment]
        assignsource: Source = self.parser.parse_note(expr, only=Source)  # type: ignore[arg-type]
        assigns.set(assigntarget, assignsource, cast=cast, overwrite=overwrite)

    def get_instcons(self, inst: Union["BaseMod", str]) -> Assigns:
        """Retrieve All Instance Connections Of `inst`."""
        mod: BaseMod = self.get_inst(inst)
        return self.__instcons[mod.name][0]

    def add_flipflop(
        self,
        type_: BaseType,
        name: str,
        clk: Parseable,
        rst_an: Parseable,
        nxt: Parseable | None = None,
        rst: Parseable | None = None,
        ena: Parseable | None = None,
        route: Routeables | None = None,
    ) -> Signal:
        """
        Add Module Internal Flip-Flop.

        Args:
            type_: Type.
            name: Name.
            clk: Clock. Module Clock by default.
            rst_an: Reset. Module Reset by default.

        Keyword Args:
            nxt: Next Value. Basename of `name` with _nxt_s by default.
            rst: Synchronous Reset.
            ena: Enable Condition.
            route: Routing of flip-flop output.
        """
        parser = self.parser
        if self.__is_locked:
            raise LockError(f"{self}: Cannot add flipflop {name!r}.")
        out = self.add_signal(type_, name)
        # clk
        clk_sig: BaseSignal = parser.parse(clk, only=BaseSignal)  # type: ignore[assignment]
        # rst_an
        rst_an_sig: BaseSignal = parser.parse(rst_an, only=BaseSignal)  # type: ignore[assignment]
        # nxt
        if nxt is None:
            nxt = self.add_signal(type_, f"{out.basename}_nxt_s")
        else:
            nxt = parser.parse(nxt)
            # TODO: check connectable of nxt and out?
        # rst
        rst_expr: BoolOp | None = cast_booltype(parser.parse(rst)) if rst is not None else None
        # ena
        ena_expr: BoolOp | None = cast_booltype(parser.parse(ena)) if ena is not None else None
        # flipflop
        flipflop = self._create_flipflop(clk_sig, rst_an_sig, rst_expr, ena_expr)
        flipflop.set(out, nxt)
        # route
        for routepath in parse_routepaths(route):
            self._router.add(RoutePath(expr=out), routepath)
        return out

    def _create_flipflop(
        self,
        clk: BaseSignal,
        rst_an: BaseSignal,
        rst: BoolOp | None = None,
        ena: BoolOp | None = None,
    ) -> FlipFlop:
        flipflops = self.__flipflops
        key = hash((clk, rst_an, rst, ena))
        try:
            return flipflops[key]
        except KeyError:
            pass
        flipflops[key] = flipflop = FlipFlop(
            clk=clk,
            rst_an=rst_an,
            rst=rst,
            ena=ena,
            targets=self.portssignals,
            sources=self.namespace,
            drivers=self.drivers,
        )
        return flipflop

    @property
    def flipflops(self) -> tuple[FlipFlop, ...]:
        """
        Flip Flops.
        """
        return tuple(self.__flipflops.values())

    def add_mux(
        self,
        name,
        title: str | None = None,
        descr: str | None = None,
        comment: str | None = None,
    ) -> Mux:
        """
        Add Multiplexer with `name` And Return It For Filling.

        Args:
            name (str): Name.

        Keyword Args:
            title (str): Full Spoken Name.
            descr (str): Documentation Description.
            comment (str): Source Code Comment.

        See :any:`Mux.set()` how to fill the multiplexer and the example above.
        """
        if self.__is_locked:
            raise LockError(f"{self}: Cannot add mux {name!r}.")
        doc = Doc(title=title, descr=descr, comment=comment)
        self.__muxes[name] = mux = Mux(
            name=name,
            targets=self.portssignals,
            namespace=self.namespace,
            # drivers=self.drivers,
            parser=self.parser,
            doc=doc,
        )
        return mux

    @property
    def muxes(self) -> tuple[Mux, ...]:
        """
        Iterate over all Multiplexer.
        """
        return tuple(self.__muxes.values())

    def get_mux(self, mux: Mux | str) -> Mux:
        """Get Multiplexer."""
        if not isinstance(mux, str):
            return self.__muxes.get_dym(mux.name)  # type: ignore[return-value]
        return self.__muxes.get_dym(mux)  # type: ignore[return-value]

    @property
    def is_locked(self) -> bool:
        """
        Return If Module Is Already Completed And Locked For Modification.

        Locking is done by the build process **automatically** and **MUST NOT** be done earlier or later.
        Use a different module type or enumeration or struct type, if you have issues with locking.
        """
        return self.__is_locked

    def lock(self):
        """
        Lock.

        Locking is done via this method by the build process **automatically** and **MUST NOT** be done earlier or
        later. Use a different module type or enumeration or struct type, if you have issues with locking.
        """
        if self.__is_locked:
            raise LockError(f"{self} is already locked. Cannot lock again.")
        for _, obj in self:
            if isinstance(obj, Namespace):
                obj.lock()
        self.__is_locked = True

    def check_lock(self):
        """Check if module is locked for modifications."""
        if self.__is_locked:
            raise LockError(f"{self}: Is already locked for modifications.")

    def con(self, port: Routeables, source: Routeables, on_error: RoutingError = "error"):
        """Connect `port` to `dest`."""
        parents = self.__parents
        if not parents:
            raise TypeError(f"{self} is top module. Connections cannot be made.")
        router = parents[-1]._router
        for subtarget in parse_routepaths(port, basepath=self.name):
            for subsource in parse_routepaths(source):
                router.add(subtarget, subsource, on_error=on_error)

    def route(self, target: Routeables, source: Routeables, on_error: RoutingError = "error"):
        """Route `source` to `target` within the actual module."""
        router = self._router
        for subtarget in parse_routepaths(target):
            for subsource in parse_routepaths(source):
                router.add(subtarget, subsource, on_error=on_error)

    def __str__(self):
        modref = self.get_modref()
        defines = ""
        if self.defines:
            definesdict = {define.name: define.value for define in self.defines}
            defines = f" defines={definesdict!r}"
        return f"<{modref}(inst={self.inst!r}, libname={self.libname!r}, modname={self.modname!r}{defines})>"

    def __repr__(self):
        modref = self.get_modref()
        defines = ""
        if self.defines:
            definesdict = {define.name: define.value for define in self.defines}
            defines = f" defines={definesdict!r}"
        return f"<{modref}(inst={self.inst!r}, libname={self.libname!r}, modname={self.modname!r}{defines})>"

    def get_overview(self) -> str:
        """
        Return Brief Module Overview.

        This Module Overview is intended to be overwritten by the user.
        """
        return ""

    def get_info(self, sub: bool = False) -> str:
        """Module Information."""
        header = f"## `{self.libname}.{self.modname}` (`{self.get_modref()}`)"
        parts = [
            header,
            self._get_ident_info("Parameters", self.params),
            self._get_ident_info("Ports", self.ports),
        ]
        if sub:
            parts.append(self._get_sub_info())
        return "\n\n".join(parts)

    def _get_ident_info(self, title: str, idents: Idents):
        def entry(level, ident):
            pre = "  " * level
            dinfo = f" ({ident.direction})" if ident.direction else ""
            return (
                f"{pre}{ident.name}{dinfo}",
                f"{pre}{ident.type_}",
            )

        parts = [
            f"### {title}",
            "",
        ]
        if idents:
            data = [("Name ", "Type"), ("----", "----")]
            data += [entry(level, ident) for level, ident in idents.leveliter()]
            parts.append(align(data, seps=(" | ", " |"), sepfirst="| "))
        else:
            parts.append("-")
        return "\n".join(parts)

    def _get_sub_info(self) -> str:
        parts = [
            "### Submodules",
            "",
        ]
        if self.insts:
            data = [("Name", "Module"), ("----", "------")]
            data += [(f"`{inst.name}`", f"`{inst.libname}.{inst.modname}`") for inst in self.insts]
            parts.append(align(data, seps=(" | ", " |"), sepfirst="| "))
        else:
            parts.append("-")
        return "\n".join(parts)


class Router(Object):
    """The One And Only Router."""

    mod: BaseMod
    __routes: list[tuple[RoutePath, RoutePath, RoutingError]] = PrivateField(default_factory=list)

    def add(self, tpath: RoutePath, spath: RoutePath, on_error: RoutingError = "error") -> None:
        """Add route from `source` to `tpath`."""
        LOGGER.debug("%s: router: add '%s' to '%s'", self.mod, spath, tpath)
        self.__routes.append(self._create(tpath, spath, on_error))

    def flush(self) -> None:
        """Create Pending Routes."""
        for tpath, spath, on_error in self.__routes:
            tpathc, spathc, on_errorc = self._create(tpath, spath, on_error)
            try:
                self._route(tpathc, spathc)
            except Exception as exc:
                if on_errorc == "ignore":
                    LOGGER.info("Ignored: %s", exc)
                elif on_errorc == "warn":
                    LOGGER.warning(exc)
                else:
                    raise
        self.__routes.clear()

    def _create(
        self, tpath: RoutePath, spath: RoutePath, on_error: RoutingError
    ) -> tuple[RoutePath, RoutePath, RoutingError]:
        if tpath.create:
            if self.__create(spath, tpath):
                tpath = tpath.new(create=False)
        elif spath.create:
            if self.__create(tpath, spath):
                spath = spath.new(create=False)
        return tpath, spath, on_error

    @no_type_check  # TODO: fix types
    def __create(self, rpath: RoutePath, cpath: RoutePath) -> bool:
        """Create `cpath` based on `rpath`."""
        assert not rpath.create
        assert cpath.create
        mod = self.mod
        # Resolve reference path
        try:
            rmod = mod.get_inst(rpath.path) if rpath.path else mod
            rident: Ident = rmod.parser.parse(rpath.expr, only=Ident)  # type: ignore[assignment]
            cmod = mod.get_inst(cpath.path) if cpath.path else mod
        except (ValueError, NameError, KeyError):
            return False
        self.__create_port_or_signal(cmod, rident, cpath.expr)
        return True

    @no_type_check  # TODO: fix types
    def _route(self, tpath: RoutePath, spath: RoutePath):  # noqa: C901, PLR0912, PLR0915
        mod = self.mod
        LOGGER.debug("%s router: routing %r to %r", mod, spath, tpath)
        # Referenced modules
        tmod = mod.get_inst(tpath.path) if tpath.path else mod
        smod = mod.get_inst(spath.path) if spath.path else mod
        # Referenced expression/signal
        texpr = tmod.parser.parse(tpath.expr) if not isinstance(tpath.expr, Note) else tpath.expr
        sexpr = smod.parser.parse(spath.expr) if not isinstance(spath.expr, Note) else spath.expr
        tident = None if not isinstance(texpr, Ident) else texpr
        sident = None if not isinstance(sexpr, Ident) else sexpr
        tparts = tpath.parts
        sparts = spath.parts
        # One of the both sides need to exist
        rident = tident or sident
        assert rident is not None
        direction = (
            tident.direction * sident.direction
            if tident and tident.direction is not None and sident and sident.direction is not None
            else rident.direction
        )

        cast = _merge_cast(tpath.cast, spath.cast)
        assert len(tparts) in (0, 1)
        assert len(sparts) in (0, 1)
        if tparts:
            # target is submodule
            assert tparts[0] != UPWARDS
            if sparts:
                assert sparts[0] != UPWARDS
                # source and target are submodules
                tcon = None if tident is None else mod.get_instcons(tmod).get(tident)
                scon = None if sident is None else mod.get_instcons(smod).get(sident)
                if tcon is None and scon is None:
                    modname = split_prefix(tmod.name)[1]
                    name = join_names(modname, rident.name, "s")
                    rsig = mod.add_signal(rident.type_, name, ifdefs=rident.ifdefs, direction=direction)
                    Router.__routesubmod(mod, tmod, rident, texpr, rsig, tname=tpath.expr, cast=tpath.cast)
                    Router.__routesubmod(mod, smod, rident, sexpr, rsig, tname=spath.expr, cast=spath.cast)
                elif tcon is None:
                    Router.__routesubmod(mod, tmod, rident, texpr, scon, tname=tpath.expr, cast=tpath.cast)
                elif scon is None:
                    Router.__routesubmod(mod, smod, rident, sexpr, tcon, tname=spath.expr, cast=spath.cast)
                else:
                    mod.assign(tcon, scon, cast=cast)
            else:
                tcon = None if tident is None else mod.get_instcons(tmod).get(tident)
                if tcon is None:
                    Router.__routesubmod(mod, tmod, rident, texpr, sexpr, tpath.expr, spath.expr, cast=cast)
                else:
                    if sexpr is None:
                        sexpr = Router.__create_port_or_signal(mod, rident, spath.expr)
                    mod.assign(tcon, sexpr, cast=cast)
        elif sparts:
            scon = None if sident is None else mod.get_instcons(smod).get(sident)
            assert sparts[0] != UPWARDS
            if scon is None:
                Router.__routesubmod(mod, smod, rident, sexpr, texpr, spath.expr, tpath.expr, cast=cast)
            else:
                if texpr is None:
                    texpr = Router.__create_port_or_signal(mod, rident, tpath.expr)
                mod.assign(scon, texpr, cast=cast)
        else:
            # connect signals of `mod`
            if texpr is None:
                texpr = Router.__create_port_or_signal(mod, rident, tpath.expr)
            if sexpr is None:
                sexpr = Router.__create_port_or_signal(mod, rident, spath.expr)
            mod.assign(texpr, sexpr, cast=cast)

    @staticmethod
    def __routesubmod(
        mod: BaseMod, submod: BaseMod, rident: Ident, texpr, sexpr, tname=None, sname=None, cast=False
    ) -> Expr:
        if texpr is None:
            assert tname is not None
            assert rident is not None and rident.type_
            texpr = submod.add_port(rident.type_, tname, ifdefs=rident.ifdefs)
        if sexpr is None:
            sexpr = Router.__create_port_or_signal(mod, rident, sname)
        if not isinstance(texpr, Port):
            raise ValueError(f"Cannot route {type(texpr)} to module instance {submod}")
        try:
            mod.set_instcon(submod, texpr, sexpr, cast=cast)
        except TypeError as err:
            raise TypeError(f"{mod}: {err}") from None
        return sexpr

    @staticmethod
    def __create_port_or_signal(mod: BaseMod, rident: Ident, name: str) -> BaseSignal:
        assert isinstance(rident, Ident)
        assert name is not None
        assert rident is not None and rident.type_ is not None, (mod, name)
        type_ = rident.type_
        direction = Direction.from_name(name)
        signal: BaseSignal
        if direction is not None:
            signal = mod.add_port(type_, name, ifdefs=rident.ifdefs, direction=direction)
        else:
            signal = mod.add_signal(type_, name, ifdefs=rident.ifdefs)
        LOGGER.debug("%s: router: creating %r", mod, signal)
        return signal


def _merge_cast(one, other):
    # TODO: get rid of this.
    if one or other:
        return True
    if one is None or other is None:
        return None
    return False


ModCls: TypeAlias = type[BaseMod]
ModClss: TypeAlias = set[type[BaseMod]]


def get_modbaseclss(cls):
    """Get Module Base Classes."""
    clss = []
    for basecls in getmro(cls):
        if basecls is BaseMod:
            break
        clss.append(basecls)
    return clss
